import logging
import os
import unittest
from io import BytesIO

import pytest
from PIL import Image, TiffImagePlugin
from PIL.TiffImagePlugin import RESOLUTION_UNIT, X_RESOLUTION, Y_RESOLUTION

from .helper import PillowTestCase, hopper, is_pypy, is_win32

logger = logging.getLogger(__name__)


class TestFileTiff(PillowTestCase):
    def test_sanity(self):

        filename = self.tempfile("temp.tif")

        hopper("RGB").save(filename)

        with Image.open(filename) as im:
            im.load()
        self.assertEqual(im.mode, "RGB")
        self.assertEqual(im.size, (128, 128))
        self.assertEqual(im.format, "TIFF")

        hopper("1").save(filename)
        with Image.open(filename):
            pass

        hopper("L").save(filename)
        with Image.open(filename):
            pass

        hopper("P").save(filename)
        with Image.open(filename):
            pass

        hopper("RGB").save(filename)
        with Image.open(filename):
            pass

        hopper("I").save(filename)
        with Image.open(filename):
            pass

    @unittest.skipIf(is_pypy(), "Requires CPython")
    def test_unclosed_file(self):
        def open():
            im = Image.open("Tests/images/multipage.tiff")
            im.load()

        self.assert_warning(ResourceWarning, open)

    def test_closed_file(self):
        def open():
            im = Image.open("Tests/images/multipage.tiff")
            im.load()
            im.close()

        self.assert_warning(None, open)

    def test_context_manager(self):
        def open():
            with Image.open("Tests/images/multipage.tiff") as im:
                im.load()

        self.assert_warning(None, open)

    def test_mac_tiff(self):
        # Read RGBa images from macOS [@PIL136]

        filename = "Tests/images/pil136.tiff"
        with Image.open(filename) as im:
            self.assertEqual(im.mode, "RGBA")
            self.assertEqual(im.size, (55, 43))
            self.assertEqual(im.tile, [("raw", (0, 0, 55, 43), 8, ("RGBa", 0, 1))])
            im.load()

            self.assert_image_similar_tofile(im, "Tests/images/pil136.png", 1)

    def test_wrong_bits_per_sample(self):
        with Image.open("Tests/images/tiff_wrong_bits_per_sample.tiff") as im:
            self.assertEqual(im.mode, "RGBA")
            self.assertEqual(im.size, (52, 53))
            self.assertEqual(im.tile, [("raw", (0, 0, 52, 53), 160, ("RGBA", 0, 1))])
            im.load()

    def test_set_legacy_api(self):
        ifd = TiffImagePlugin.ImageFileDirectory_v2()
        with self.assertRaises(Exception) as e:
            ifd.legacy_api = None
        self.assertEqual(str(e.exception), "Not allowing setting of legacy api")

    def test_xyres_tiff(self):
        filename = "Tests/images/pil168.tif"
        with Image.open(filename) as im:

            # legacy api
            self.assertIsInstance(im.tag[X_RESOLUTION][0], tuple)
            self.assertIsInstance(im.tag[Y_RESOLUTION][0], tuple)

            # v2 api
            self.assertIsInstance(im.tag_v2[X_RESOLUTION], TiffImagePlugin.IFDRational)
            self.assertIsInstance(im.tag_v2[Y_RESOLUTION], TiffImagePlugin.IFDRational)

            self.assertEqual(im.info["dpi"], (72.0, 72.0))

    def test_xyres_fallback_tiff(self):
        filename = "Tests/images/compression.tif"
        with Image.open(filename) as im:

            # v2 api
            self.assertIsInstance(im.tag_v2[X_RESOLUTION], TiffImagePlugin.IFDRational)
            self.assertIsInstance(im.tag_v2[Y_RESOLUTION], TiffImagePlugin.IFDRational)
            self.assertRaises(KeyError, lambda: im.tag_v2[RESOLUTION_UNIT])

            # Legacy.
            self.assertEqual(im.info["resolution"], (100.0, 100.0))
            # Fallback "inch".
            self.assertEqual(im.info["dpi"], (100.0, 100.0))

    def test_int_resolution(self):
        filename = "Tests/images/pil168.tif"
        with Image.open(filename) as im:

            # Try to read a file where X,Y_RESOLUTION are ints
            im.tag_v2[X_RESOLUTION] = 71
            im.tag_v2[Y_RESOLUTION] = 71
            im._setup()
            self.assertEqual(im.info["dpi"], (71.0, 71.0))

    def test_load_dpi_rounding(self):
        for resolutionUnit, dpi in ((None, (72, 73)), (2, (72, 73)), (3, (183, 185))):
            with Image.open(
                "Tests/images/hopper_roundDown_" + str(resolutionUnit) + ".tif"
            ) as im:
                self.assertEqual(im.tag_v2.get(RESOLUTION_UNIT), resolutionUnit)
                self.assertEqual(im.info["dpi"], (dpi[0], dpi[0]))

            with Image.open(
                "Tests/images/hopper_roundUp_" + str(resolutionUnit) + ".tif"
            ) as im:
                self.assertEqual(im.tag_v2.get(RESOLUTION_UNIT), resolutionUnit)
                self.assertEqual(im.info["dpi"], (dpi[1], dpi[1]))

    def test_save_dpi_rounding(self):
        outfile = self.tempfile("temp.tif")
        with Image.open("Tests/images/hopper.tif") as im:
            for dpi in (72.2, 72.8):
                im.save(outfile, dpi=(dpi, dpi))

                with Image.open(outfile) as reloaded:
                    reloaded.load()
                    self.assertEqual((round(dpi), round(dpi)), reloaded.info["dpi"])

    def test_save_setting_missing_resolution(self):
        b = BytesIO()
        Image.open("Tests/images/10ct_32bit_128.tiff").save(
            b, format="tiff", resolution=123.45
        )
        with Image.open(b) as im:
            self.assertEqual(float(im.tag_v2[X_RESOLUTION]), 123.45)
            self.assertEqual(float(im.tag_v2[Y_RESOLUTION]), 123.45)

    def test_invalid_file(self):
        invalid_file = "Tests/images/flower.jpg"

        self.assertRaises(SyntaxError, TiffImagePlugin.TiffImageFile, invalid_file)

        TiffImagePlugin.PREFIXES.append(b"\xff\xd8\xff\xe0")
        self.assertRaises(SyntaxError, TiffImagePlugin.TiffImageFile, invalid_file)
        TiffImagePlugin.PREFIXES.pop()

    def test_bad_exif(self):
        with Image.open("Tests/images/hopper_bad_exif.jpg") as i:
            # Should not raise struct.error.
            self.assert_warning(UserWarning, i._getexif)

    def test_save_rgba(self):
        im = hopper("RGBA")
        outfile = self.tempfile("temp.tif")
        im.save(outfile)

    def test_save_unsupported_mode(self):
        im = hopper("HSV")
        outfile = self.tempfile("temp.tif")
        self.assertRaises(IOError, im.save, outfile)

    def test_little_endian(self):
        with Image.open("Tests/images/16bit.cropped.tif") as im:
            self.assertEqual(im.getpixel((0, 0)), 480)
            self.assertEqual(im.mode, "I;16")

            b = im.tobytes()
        # Bytes are in image native order (little endian)
        self.assertEqual(b[0], ord(b"\xe0"))
        self.assertEqual(b[1], ord(b"\x01"))

    def test_big_endian(self):
        with Image.open("Tests/images/16bit.MM.cropped.tif") as im:
            self.assertEqual(im.getpixel((0, 0)), 480)
            self.assertEqual(im.mode, "I;16B")

            b = im.tobytes()
        # Bytes are in image native order (big endian)
        self.assertEqual(b[0], ord(b"\x01"))
        self.assertEqual(b[1], ord(b"\xe0"))

    def test_16bit_s(self):
        with Image.open("Tests/images/16bit.s.tif") as im:
            im.load()
            self.assertEqual(im.mode, "I")
            self.assertEqual(im.getpixel((0, 0)), 32767)
            self.assertEqual(im.getpixel((0, 1)), 0)

    def test_12bit_rawmode(self):
        """ Are we generating the same interpretation
        of the image as Imagemagick is? """

        with Image.open("Tests/images/12bit.cropped.tif") as im:
            # to make the target --
            # convert 12bit.cropped.tif -depth 16 tmp.tif
            # convert tmp.tif -evaluate RightShift 4 12in16bit2.tif
            # imagemagick will auto scale so that a 12bit FFF is 16bit FFF0,
            # so we need to unshift so that the integer values are the same.

            self.assert_image_equal_tofile(im, "Tests/images/12in16bit.tif")

    def test_32bit_float(self):
        # Issue 614, specific 32-bit float format
        path = "Tests/images/10ct_32bit_128.tiff"
        with Image.open(path) as im:
            im.load()

            self.assertEqual(im.getpixel((0, 0)), -0.4526388943195343)
            self.assertEqual(im.getextrema(), (-3.140936851501465, 3.140684127807617))

    def test_unknown_pixel_mode(self):
        self.assertRaises(
            IOError, Image.open, "Tests/images/hopper_unknown_pixel_mode.tif"
        )

    def test_n_frames(self):
        for path, n_frames in [
            ["Tests/images/multipage-lastframe.tif", 1],
            ["Tests/images/multipage.tiff", 3],
        ]:
            with Image.open(path) as im:
                self.assertEqual(im.n_frames, n_frames)
                self.assertEqual(im.is_animated, n_frames != 1)

    def test_eoferror(self):
        with Image.open("Tests/images/multipage-lastframe.tif") as im:
            n_frames = im.n_frames

            # Test seeking past the last frame
            self.assertRaises(EOFError, im.seek, n_frames)
            self.assertLess(im.tell(), n_frames)

            # Test that seeking to the last frame does not raise an error
            im.seek(n_frames - 1)

    def test_multipage(self):
        # issue #862
        with Image.open("Tests/images/multipage.tiff") as im:
            # file is a multipage tiff: 10x10 green, 10x10 red, 20x20 blue

            im.seek(0)
            self.assertEqual(im.size, (10, 10))
            self.assertEqual(im.convert("RGB").getpixel((0, 0)), (0, 128, 0))

            im.seek(1)
            im.load()
            self.assertEqual(im.size, (10, 10))
            self.assertEqual(im.convert("RGB").getpixel((0, 0)), (255, 0, 0))

            im.seek(0)
            im.load()
            self.assertEqual(im.size, (10, 10))
            self.assertEqual(im.convert("RGB").getpixel((0, 0)), (0, 128, 0))

            im.seek(2)
            im.load()
            self.assertEqual(im.size, (20, 20))
            self.assertEqual(im.convert("RGB").getpixel((0, 0)), (0, 0, 255))

    def test_multipage_last_frame(self):
        with Image.open("Tests/images/multipage-lastframe.tif") as im:
            im.load()
            self.assertEqual(im.size, (20, 20))
            self.assertEqual(im.convert("RGB").getpixel((0, 0)), (0, 0, 255))

    def test___str__(self):
        filename = "Tests/images/pil136.tiff"
        with Image.open(filename) as im:

            # Act
            ret = str(im.ifd)

            # Assert
            self.assertIsInstance(ret, str)

    def test_dict(self):
        # Arrange
        filename = "Tests/images/pil136.tiff"
        with Image.open(filename) as im:

            # v2 interface
            v2_tags = {
                256: 55,
                257: 43,
                258: (8, 8, 8, 8),
                259: 1,
                262: 2,
                296: 2,
                273: (8,),
                338: (1,),
                277: 4,
                279: (9460,),
                282: 72.0,
                283: 72.0,
                284: 1,
            }
            self.assertEqual(dict(im.tag_v2), v2_tags)

            # legacy interface
            legacy_tags = {
                256: (55,),
                257: (43,),
                258: (8, 8, 8, 8),
                259: (1,),
                262: (2,),
                296: (2,),
                273: (8,),
                338: (1,),
                277: (4,),
                279: (9460,),
                282: ((720000, 10000),),
                283: ((720000, 10000),),
                284: (1,),
            }
            self.assertEqual(dict(im.tag), legacy_tags)

    def test__delitem__(self):
        filename = "Tests/images/pil136.tiff"
        with Image.open(filename) as im:
            len_before = len(dict(im.ifd))
            del im.ifd[256]
            len_after = len(dict(im.ifd))
            self.assertEqual(len_before, len_after + 1)

    def test_load_byte(self):
        for legacy_api in [False, True]:
            ifd = TiffImagePlugin.ImageFileDirectory_v2()
            data = b"abc"
            ret = ifd.load_byte(data, legacy_api)
            self.assertEqual(ret, b"abc")

    def test_load_string(self):
        ifd = TiffImagePlugin.ImageFileDirectory_v2()
        data = b"abc\0"
        ret = ifd.load_string(data, False)
        self.assertEqual(ret, "abc")

    def test_load_float(self):
        ifd = TiffImagePlugin.ImageFileDirectory_v2()
        data = b"abcdabcd"
        ret = ifd.load_float(data, False)
        self.assertEqual(ret, (1.6777999408082104e22, 1.6777999408082104e22))

    def test_load_double(self):
        ifd = TiffImagePlugin.ImageFileDirectory_v2()
        data = b"abcdefghabcdefgh"
        ret = ifd.load_double(data, False)
        self.assertEqual(ret, (8.540883223036124e194, 8.540883223036124e194))

    def test_seek(self):
        filename = "Tests/images/pil136.tiff"
        with Image.open(filename) as im:
            im.seek(0)
            self.assertEqual(im.tell(), 0)

    def test_seek_eof(self):
        filename = "Tests/images/pil136.tiff"
        with Image.open(filename) as im:
            self.assertEqual(im.tell(), 0)
            self.assertRaises(EOFError, im.seek, -1)
            self.assertRaises(EOFError, im.seek, 1)

    def test__limit_rational_int(self):
        from PIL.TiffImagePlugin import _limit_rational

        value = 34
        ret = _limit_rational(value, 65536)
        self.assertEqual(ret, (34, 1))

    def test__limit_rational_float(self):
        from PIL.TiffImagePlugin import _limit_rational

        value = 22.3
        ret = _limit_rational(value, 65536)
        self.assertEqual(ret, (223, 10))

    def test_4bit(self):
        test_file = "Tests/images/hopper_gray_4bpp.tif"
        original = hopper("L")
        with Image.open(test_file) as im:
            self.assertEqual(im.size, (128, 128))
            self.assertEqual(im.mode, "L")
            self.assert_image_similar(im, original, 7.3)

    def test_gray_semibyte_per_pixel(self):
        test_files = (
            (
                24.8,  # epsilon
                (  # group
                    "Tests/images/tiff_gray_2_4_bpp/hopper2.tif",
                    "Tests/images/tiff_gray_2_4_bpp/hopper2I.tif",
                    "Tests/images/tiff_gray_2_4_bpp/hopper2R.tif",
                    "Tests/images/tiff_gray_2_4_bpp/hopper2IR.tif",
                ),
            ),
            (
                7.3,  # epsilon
                (  # group
                    "Tests/images/tiff_gray_2_4_bpp/hopper4.tif",
                    "Tests/images/tiff_gray_2_4_bpp/hopper4I.tif",
                    "Tests/images/tiff_gray_2_4_bpp/hopper4R.tif",
                    "Tests/images/tiff_gray_2_4_bpp/hopper4IR.tif",
                ),
            ),
        )
        original = hopper("L")
        for epsilon, group in test_files:
            with Image.open(group[0]) as im:
                self.assertEqual(im.size, (128, 128))
                self.assertEqual(im.mode, "L")
                self.assert_image_similar(im, original, epsilon)
                for file in group[1:]:
                    with Image.open(file) as im2:
                        self.assertEqual(im2.size, (128, 128))
                        self.assertEqual(im2.mode, "L")
                        self.assert_image_equal(im, im2)

    def test_with_underscores(self):
        kwargs = {"resolution_unit": "inch", "x_resolution": 72, "y_resolution": 36}
        filename = self.tempfile("temp.tif")
        hopper("RGB").save(filename, **kwargs)
        with Image.open(filename) as im:

            # legacy interface
            self.assertEqual(im.tag[X_RESOLUTION][0][0], 72)
            self.assertEqual(im.tag[Y_RESOLUTION][0][0], 36)

            # v2 interface
            self.assertEqual(im.tag_v2[X_RESOLUTION], 72)
            self.assertEqual(im.tag_v2[Y_RESOLUTION], 36)

    def test_roundtrip_tiff_uint16(self):
        # Test an image of all '0' values
        pixel_value = 0x1234
        infile = "Tests/images/uint16_1_4660.tif"
        with Image.open(infile) as im:
            self.assertEqual(im.getpixel((0, 0)), pixel_value)

            tmpfile = self.tempfile("temp.tif")
            im.save(tmpfile)

            with Image.open(tmpfile) as reloaded:
                self.assert_image_equal(im, reloaded)

    def test_strip_raw(self):
        infile = "Tests/images/tiff_strip_raw.tif"
        with Image.open(infile) as im:
            self.assert_image_equal_tofile(im, "Tests/images/tiff_adobe_deflate.png")

    def test_strip_planar_raw(self):
        # gdal_translate -of GTiff -co INTERLEAVE=BAND \
        # tiff_strip_raw.tif tiff_strip_planar_raw.tiff
        infile = "Tests/images/tiff_strip_planar_raw.tif"
        with Image.open(infile) as im:
            self.assert_image_equal_tofile(im, "Tests/images/tiff_adobe_deflate.png")

    def test_strip_planar_raw_with_overviews(self):
        # gdaladdo tiff_strip_planar_raw2.tif 2 4 8 16
        infile = "Tests/images/tiff_strip_planar_raw_with_overviews.tif"
        with Image.open(infile) as im:
            self.assert_image_equal_tofile(im, "Tests/images/tiff_adobe_deflate.png")

    def test_tiled_planar_raw(self):
        # gdal_translate -of GTiff -co TILED=YES -co BLOCKXSIZE=32 \
        # -co BLOCKYSIZE=32 -co INTERLEAVE=BAND \
        # tiff_tiled_raw.tif tiff_tiled_planar_raw.tiff
        infile = "Tests/images/tiff_tiled_planar_raw.tif"
        with Image.open(infile) as im:
            self.assert_image_equal_tofile(im, "Tests/images/tiff_adobe_deflate.png")

    def test_palette(self):
        for mode in ["P", "PA"]:
            outfile = self.tempfile("temp.tif")

            im = hopper(mode)
            im.save(outfile)

            with Image.open(outfile) as reloaded:
                self.assert_image_equal(im.convert("RGB"), reloaded.convert("RGB"))

    def test_tiff_save_all(self):
        mp = BytesIO()
        with Image.open("Tests/images/multipage.tiff") as im:
            im.save(mp, format="tiff", save_all=True)

        mp.seek(0, os.SEEK_SET)
        with Image.open(mp) as im:
            self.assertEqual(im.n_frames, 3)

        # Test appending images
        mp = BytesIO()
        im = Image.new("RGB", (100, 100), "#f00")
        ims = [Image.new("RGB", (100, 100), color) for color in ["#0f0", "#00f"]]
        im.copy().save(mp, format="TIFF", save_all=True, append_images=ims)

        mp.seek(0, os.SEEK_SET)
        with Image.open(mp) as reread:
            self.assertEqual(reread.n_frames, 3)

        # Test appending using a generator
        def imGenerator(ims):
            yield from ims

        mp = BytesIO()
        im.save(mp, format="TIFF", save_all=True, append_images=imGenerator(ims))

        mp.seek(0, os.SEEK_SET)
        with Image.open(mp) as reread:
            self.assertEqual(reread.n_frames, 3)

    def test_saving_icc_profile(self):
        # Tests saving TIFF with icc_profile set.
        # At the time of writing this will only work for non-compressed tiffs
        # as libtiff does not support embedded ICC profiles,
        # ImageFile._save(..) however does.
        im = Image.new("RGB", (1, 1))
        im.info["icc_profile"] = "Dummy value"

        # Try save-load round trip to make sure both handle icc_profile.
        tmpfile = self.tempfile("temp.tif")
        im.save(tmpfile, "TIFF", compression="raw")
        with Image.open(tmpfile) as reloaded:
            self.assertEqual(b"Dummy value", reloaded.info["icc_profile"])

    def test_close_on_load_exclusive(self):
        # similar to test_fd_leak, but runs on unixlike os
        tmpfile = self.tempfile("temp.tif")

        with Image.open("Tests/images/uint16_1_4660.tif") as im:
            im.save(tmpfile)

        im = Image.open(tmpfile)
        fp = im.fp
        self.assertFalse(fp.closed)
        im.load()
        self.assertTrue(fp.closed)

    def test_close_on_load_nonexclusive(self):
        tmpfile = self.tempfile("temp.tif")

        with Image.open("Tests/images/uint16_1_4660.tif") as im:
            im.save(tmpfile)

        with open(tmpfile, "rb") as f:
            im = Image.open(f)
            fp = im.fp
            self.assertFalse(fp.closed)
            im.load()
            self.assertFalse(fp.closed)

    # Ignore this UserWarning which triggers for four tags:
    # "Possibly corrupt EXIF data.  Expecting to read 50404352 bytes but..."
    @pytest.mark.filterwarnings("ignore:Possibly corrupt EXIF data")
    def test_string_dimension(self):
        # Assert that an error is raised if one of the dimensions is a string
        with self.assertRaises(ValueError):
            Image.open("Tests/images/string_dimension.tiff")


@unittest.skipUnless(is_win32(), "Windows only")
class TestFileTiffW32(PillowTestCase):
    def test_fd_leak(self):
        tmpfile = self.tempfile("temp.tif")

        # this is an mmaped file.
        with Image.open("Tests/images/uint16_1_4660.tif") as im:
            im.save(tmpfile)

        im = Image.open(tmpfile)
        fp = im.fp
        self.assertFalse(fp.closed)
        self.assertRaises(WindowsError, os.remove, tmpfile)
        im.load()
        self.assertTrue(fp.closed)

        # this closes the mmap
        im.close()

        # this should not fail, as load should have closed the file pointer,
        # and close should have closed the mmap
        os.remove(tmpfile)



import unittest
import zlib
from io import BytesIO

from PIL import Image, ImageFile, PngImagePlugin

from .helper import PillowLeakTestCase, PillowTestCase, hopper, is_win32

try:
    from PIL import _webp

    HAVE_WEBP = True
except ImportError:
    HAVE_WEBP = False

codecs = dir(Image.core)


# sample png stream

TEST_PNG_FILE = "Tests/images/hopper.png"

# stuff to create inline PNG images

MAGIC = PngImagePlugin._MAGIC


def chunk(cid, *data):
    test_file = BytesIO()
    PngImagePlugin.putchunk(*(test_file, cid) + data)
    return test_file.getvalue()


o32 = PngImagePlugin.o32

IHDR = chunk(b"IHDR", o32(1), o32(1), b"\x08\x02", b"\0\0\0")
IDAT = chunk(b"IDAT")
IEND = chunk(b"IEND")

HEAD = MAGIC + IHDR
TAIL = IDAT + IEND


def load(data):
    return Image.open(BytesIO(data))


def roundtrip(im, **options):
    out = BytesIO()
    im.save(out, "PNG", **options)
    out.seek(0)
    return Image.open(out)


class TestFilePng(PillowTestCase):
    def setUp(self):
        if "zip_encoder" not in codecs or "zip_decoder" not in codecs:
            self.skipTest("zip/deflate support not available")

    def get_chunks(self, filename):
        chunks = []
        with open(filename, "rb") as fp:
            fp.read(8)
            with PngImagePlugin.PngStream(fp) as png:
                while True:
                    cid, pos, length = png.read()
                    chunks.append(cid)
                    try:
                        s = png.call(cid, pos, length)
                    except EOFError:
                        break
                    png.crc(cid, s)
        return chunks

    def test_sanity(self):

        # internal version number
        self.assertRegex(Image.core.zlib_version, r"\d+\.\d+\.\d+(\.\d+)?$")

        test_file = self.tempfile("temp.png")

        hopper("RGB").save(test_file)

        with Image.open(test_file) as im:
            im.load()
            self.assertEqual(im.mode, "RGB")
            self.assertEqual(im.size, (128, 128))
            self.assertEqual(im.format, "PNG")
            self.assertEqual(im.get_format_mimetype(), "image/png")

        for mode in ["1", "L", "P", "RGB", "I", "I;16"]:
            im = hopper(mode)
            im.save(test_file)
            with Image.open(test_file) as reloaded:
                if mode == "I;16":
                    reloaded = reloaded.convert(mode)
                self.assert_image_equal(reloaded, im)

    def test_invalid_file(self):
        invalid_file = "Tests/images/flower.jpg"

        self.assertRaises(SyntaxError, PngImagePlugin.PngImageFile, invalid_file)

    def test_broken(self):
        # Check reading of totally broken files.  In this case, the test
        # file was checked into Subversion as a text file.

        test_file = "Tests/images/broken.png"
        self.assertRaises(IOError, Image.open, test_file)

    def test_bad_text(self):
        # Make sure PIL can read malformed tEXt chunks (@PIL152)

        im = load(HEAD + chunk(b"tEXt") + TAIL)
        self.assertEqual(im.info, {})

        im = load(HEAD + chunk(b"tEXt", b"spam") + TAIL)
        self.assertEqual(im.info, {"spam": ""})

        im = load(HEAD + chunk(b"tEXt", b"spam\0") + TAIL)
        self.assertEqual(im.info, {"spam": ""})

        im = load(HEAD + chunk(b"tEXt", b"spam\0egg") + TAIL)
        self.assertEqual(im.info, {"spam": "egg"})

        im = load(HEAD + chunk(b"tEXt", b"spam\0egg\0") + TAIL)
        self.assertEqual(im.info, {"spam": "egg\x00"})

    def test_bad_ztxt(self):
        # Test reading malformed zTXt chunks (python-pillow/Pillow#318)

        im = load(HEAD + chunk(b"zTXt") + TAIL)
        self.assertEqual(im.info, {})

        im = load(HEAD + chunk(b"zTXt", b"spam") + TAIL)
        self.assertEqual(im.info, {"spam": ""})

        im = load(HEAD + chunk(b"zTXt", b"spam\0") + TAIL)
        self.assertEqual(im.info, {"spam": ""})

        im = load(HEAD + chunk(b"zTXt", b"spam\0\0") + TAIL)
        self.assertEqual(im.info, {"spam": ""})

        im = load(HEAD + chunk(b"zTXt", b"spam\0\0" + zlib.compress(b"egg")[:1]) + TAIL)
        self.assertEqual(im.info, {"spam": ""})

        im = load(HEAD + chunk(b"zTXt", b"spam\0\0" + zlib.compress(b"egg")) + TAIL)
        self.assertEqual(im.info, {"spam": "egg"})

    def test_bad_itxt(self):

        im = load(HEAD + chunk(b"iTXt") + TAIL)
        self.assertEqual(im.info, {})

        im = load(HEAD + chunk(b"iTXt", b"spam") + TAIL)
        self.assertEqual(im.info, {})

        im = load(HEAD + chunk(b"iTXt", b"spam\0") + TAIL)
        self.assertEqual(im.info, {})

        im = load(HEAD + chunk(b"iTXt", b"spam\0\x02") + TAIL)
        self.assertEqual(im.info, {})

        im = load(HEAD + chunk(b"iTXt", b"spam\0\0\0foo\0") + TAIL)
        self.assertEqual(im.info, {})

        im = load(HEAD + chunk(b"iTXt", b"spam\0\0\0en\0Spam\0egg") + TAIL)
        self.assertEqual(im.info, {"spam": "egg"})
        self.assertEqual(im.info["spam"].lang, "en")
        self.assertEqual(im.info["spam"].tkey, "Spam")

        im = load(
            HEAD
            + chunk(b"iTXt", b"spam\0\1\0en\0Spam\0" + zlib.compress(b"egg")[:1])
            + TAIL
        )
        self.assertEqual(im.info, {"spam": ""})

        im = load(
            HEAD
            + chunk(b"iTXt", b"spam\0\1\1en\0Spam\0" + zlib.compress(b"egg"))
            + TAIL
        )
        self.assertEqual(im.info, {})

        im = load(
            HEAD
            + chunk(b"iTXt", b"spam\0\1\0en\0Spam\0" + zlib.compress(b"egg"))
            + TAIL
        )
        self.assertEqual(im.info, {"spam": "egg"})
        self.assertEqual(im.info["spam"].lang, "en")
        self.assertEqual(im.info["spam"].tkey, "Spam")

    def test_interlace(self):

        test_file = "Tests/images/pil123p.png"
        with Image.open(test_file) as im:
            self.assert_image(im, "P", (162, 150))
            self.assertTrue(im.info.get("interlace"))

            im.load()

        test_file = "Tests/images/pil123rgba.png"
        with Image.open(test_file) as im:
            self.assert_image(im, "RGBA", (162, 150))
            self.assertTrue(im.info.get("interlace"))

            im.load()

    def test_load_transparent_p(self):
        test_file = "Tests/images/pil123p.png"
        with Image.open(test_file) as im:
            self.assert_image(im, "P", (162, 150))
            im = im.convert("RGBA")
        self.assert_image(im, "RGBA", (162, 150))

        # image has 124 unique alpha values
        self.assertEqual(len(im.getchannel("A").getcolors()), 124)

    def test_load_transparent_rgb(self):
        test_file = "Tests/images/rgb_trns.png"
        with Image.open(test_file) as im:
            self.assertEqual(im.info["transparency"], (0, 255, 52))

            self.assert_image(im, "RGB", (64, 64))
            im = im.convert("RGBA")
        self.assert_image(im, "RGBA", (64, 64))

        # image has 876 transparent pixels
        self.assertEqual(im.getchannel("A").getcolors()[0][0], 876)

    def test_save_p_transparent_palette(self):
        in_file = "Tests/images/pil123p.png"
        with Image.open(in_file) as im:
            # 'transparency' contains a byte string with the opacity for
            # each palette entry
            self.assertEqual(len(im.info["transparency"]), 256)

            test_file = self.tempfile("temp.png")
            im.save(test_file)

        # check if saved image contains same transparency
        with Image.open(test_file) as im:
            self.assertEqual(len(im.info["transparency"]), 256)

            self.assert_image(im, "P", (162, 150))
            im = im.convert("RGBA")
        self.assert_image(im, "RGBA", (162, 150))

        # image has 124 unique alpha values
        self.assertEqual(len(im.getchannel("A").getcolors()), 124)

    def test_save_p_single_transparency(self):
        in_file = "Tests/images/p_trns_single.png"
        with Image.open(in_file) as im:
            # pixel value 164 is full transparent
            self.assertEqual(im.info["transparency"], 164)
            self.assertEqual(im.getpixel((31, 31)), 164)

            test_file = self.tempfile("temp.png")
            im.save(test_file)

        # check if saved image contains same transparency
        with Image.open(test_file) as im:
            self.assertEqual(im.info["transparency"], 164)
            self.assertEqual(im.getpixel((31, 31)), 164)
            self.assert_image(im, "P", (64, 64))
            im = im.convert("RGBA")
        self.assert_image(im, "RGBA", (64, 64))

        self.assertEqual(im.getpixel((31, 31)), (0, 255, 52, 0))

        # image has 876 transparent pixels
        self.assertEqual(im.getchannel("A").getcolors()[0][0], 876)

    def test_save_p_transparent_black(self):
        # check if solid black image with full transparency
        # is supported (check for #1838)
        im = Image.new("RGBA", (10, 10), (0, 0, 0, 0))
        self.assertEqual(im.getcolors(), [(100, (0, 0, 0, 0))])

        im = im.convert("P")
        test_file = self.tempfile("temp.png")
        im.save(test_file)

        # check if saved image contains same transparency
        with Image.open(test_file) as im:
            self.assertEqual(len(im.info["transparency"]), 256)
            self.assert_image(im, "P", (10, 10))
            im = im.convert("RGBA")
        self.assert_image(im, "RGBA", (10, 10))
        self.assertEqual(im.getcolors(), [(100, (0, 0, 0, 0))])

    def test_save_greyscale_transparency(self):
        for mode, num_transparent in {"1": 1994, "L": 559, "I": 559}.items():
            in_file = "Tests/images/" + mode.lower() + "_trns.png"
            with Image.open(in_file) as im:
                self.assertEqual(im.mode, mode)
                self.assertEqual(im.info["transparency"], 255)

                im_rgba = im.convert("RGBA")
            self.assertEqual(im_rgba.getchannel("A").getcolors()[0][0], num_transparent)

            test_file = self.tempfile("temp.png")
            im.save(test_file)

            with Image.open(test_file) as test_im:
                self.assertEqual(test_im.mode, mode)
                self.assertEqual(test_im.info["transparency"], 255)
                self.assert_image_equal(im, test_im)

            test_im_rgba = test_im.convert("RGBA")
            self.assertEqual(
                test_im_rgba.getchannel("A").getcolors()[0][0], num_transparent
            )

    def test_save_rgb_single_transparency(self):
        in_file = "Tests/images/caption_6_33_22.png"
        with Image.open(in_file) as im:
            test_file = self.tempfile("temp.png")
            im.save(test_file)

    def test_load_verify(self):
        # Check open/load/verify exception (@PIL150)

        with Image.open(TEST_PNG_FILE) as im:
            # Assert that there is no unclosed file warning
            self.assert_warning(None, im.verify)

        with Image.open(TEST_PNG_FILE) as im:
            im.load()
            self.assertRaises(RuntimeError, im.verify)

    def test_verify_struct_error(self):
        # Check open/load/verify exception (#1755)

        # offsets to test, -10: breaks in i32() in read. (IOError)
        #                  -13: breaks in crc, txt chunk.
        #                  -14: malformed chunk

        for offset in (-10, -13, -14):
            with open(TEST_PNG_FILE, "rb") as f:
                test_file = f.read()[:offset]

            with Image.open(BytesIO(test_file)) as im:
                self.assertIsNotNone(im.fp)
                self.assertRaises((IOError, SyntaxError), im.verify)

    def test_verify_ignores_crc_error(self):
        # check ignores crc errors in ancillary chunks

        chunk_data = chunk(b"tEXt", b"spam")
        broken_crc_chunk_data = chunk_data[:-1] + b"q"  # break CRC

        image_data = HEAD + broken_crc_chunk_data + TAIL
        self.assertRaises(SyntaxError, PngImagePlugin.PngImageFile, BytesIO(image_data))

        ImageFile.LOAD_TRUNCATED_IMAGES = True
        try:
            im = load(image_data)
            self.assertIsNotNone(im)
        finally:
            ImageFile.LOAD_TRUNCATED_IMAGES = False

    def test_verify_not_ignores_crc_error_in_required_chunk(self):
        # check does not ignore crc errors in required chunks

        image_data = MAGIC + IHDR[:-1] + b"q" + TAIL

        ImageFile.LOAD_TRUNCATED_IMAGES = True
        try:
            self.assertRaises(
                SyntaxError, PngImagePlugin.PngImageFile, BytesIO(image_data)
            )
        finally:
            ImageFile.LOAD_TRUNCATED_IMAGES = False

    def test_roundtrip_dpi(self):
        # Check dpi roundtripping

        with Image.open(TEST_PNG_FILE) as im:
            im = roundtrip(im, dpi=(100, 100))
        self.assertEqual(im.info["dpi"], (100, 100))

    def test_load_dpi_rounding(self):
        # Round up
        with Image.open(TEST_PNG_FILE) as im:
            self.assertEqual(im.info["dpi"], (96, 96))

        # Round down
        with Image.open("Tests/images/icc_profile_none.png") as im:
            self.assertEqual(im.info["dpi"], (72, 72))

    def test_save_dpi_rounding(self):
        with Image.open(TEST_PNG_FILE) as im:
            im = roundtrip(im, dpi=(72.2, 72.2))
        self.assertEqual(im.info["dpi"], (72, 72))

        im = roundtrip(im, dpi=(72.8, 72.8))
        self.assertEqual(im.info["dpi"], (73, 73))

    def test_roundtrip_text(self):
        # Check text roundtripping

        with Image.open(TEST_PNG_FILE) as im:
            info = PngImagePlugin.PngInfo()
            info.add_text("TXT", "VALUE")
            info.add_text("ZIP", "VALUE", zip=True)

            im = roundtrip(im, pnginfo=info)
        self.assertEqual(im.info, {"TXT": "VALUE", "ZIP": "VALUE"})
        self.assertEqual(im.text, {"TXT": "VALUE", "ZIP": "VALUE"})

    def test_roundtrip_itxt(self):
        # Check iTXt roundtripping

        im = Image.new("RGB", (32, 32))
        info = PngImagePlugin.PngInfo()
        info.add_itxt("spam", "Eggs", "en", "Spam")
        info.add_text("eggs", PngImagePlugin.iTXt("Spam", "en", "Eggs"), zip=True)

        im = roundtrip(im, pnginfo=info)
        self.assertEqual(im.info, {"spam": "Eggs", "eggs": "Spam"})
        self.assertEqual(im.text, {"spam": "Eggs", "eggs": "Spam"})
        self.assertEqual(im.text["spam"].lang, "en")
        self.assertEqual(im.text["spam"].tkey, "Spam")
        self.assertEqual(im.text["eggs"].lang, "en")
        self.assertEqual(im.text["eggs"].tkey, "Eggs")

    def test_nonunicode_text(self):
        # Check so that non-Unicode text is saved as a tEXt rather than iTXt

        im = Image.new("RGB", (32, 32))
        info = PngImagePlugin.PngInfo()
        info.add_text("Text", "Ascii")
        im = roundtrip(im, pnginfo=info)
        self.assertIsInstance(im.info["Text"], str)

    def test_unicode_text(self):
        # Check preservation of non-ASCII characters

        def rt_text(value):
            im = Image.new("RGB", (32, 32))
            info = PngImagePlugin.PngInfo()
            info.add_text("Text", value)
            im = roundtrip(im, pnginfo=info)
            self.assertEqual(im.info, {"Text": value})

        rt_text(" Aa" + chr(0xA0) + chr(0xC4) + chr(0xFF))  # Latin1
        rt_text(chr(0x400) + chr(0x472) + chr(0x4FF))  # Cyrillic
        # CJK:
        rt_text(chr(0x4E00) + chr(0x66F0) + chr(0x9FBA) + chr(0x3042) + chr(0xAC00))
        rt_text("A" + chr(0xC4) + chr(0x472) + chr(0x3042))  # Combined

    def test_scary(self):
        # Check reading of evil PNG file.  For information, see:
        # http://scary.beasts.org/security/CESA-2004-001.txt
        # The first byte is removed from pngtest_bad.png
        # to avoid classification as malware.

        with open("Tests/images/pngtest_bad.png.bin", "rb") as fd:
            data = b"\x89" + fd.read()

        pngfile = BytesIO(data)
        self.assertRaises(IOError, Image.open, pngfile)

    def test_trns_rgb(self):
        # Check writing and reading of tRNS chunks for RGB images.
        # Independent file sample provided by Sebastian Spaeth.

        test_file = "Tests/images/caption_6_33_22.png"
        with Image.open(test_file) as im:
            self.assertEqual(im.info["transparency"], (248, 248, 248))

            # check saving transparency by default
            im = roundtrip(im)
        self.assertEqual(im.info["transparency"], (248, 248, 248))

        im = roundtrip(im, transparency=(0, 1, 2))
        self.assertEqual(im.info["transparency"], (0, 1, 2))

    def test_trns_p(self):
        # Check writing a transparency of 0, issue #528
        im = hopper("P")
        im.info["transparency"] = 0

        f = self.tempfile("temp.png")
        im.save(f)

        with Image.open(f) as im2:
            self.assertIn("transparency", im2.info)

            self.assert_image_equal(im2.convert("RGBA"), im.convert("RGBA"))

    def test_trns_null(self):
        # Check reading images with null tRNS value, issue #1239
        test_file = "Tests/images/tRNS_null_1x1.png"
        with Image.open(test_file) as im:

            self.assertEqual(im.info["transparency"], 0)

    def test_save_icc_profile(self):
        with Image.open("Tests/images/icc_profile_none.png") as im:
            self.assertIsNone(im.info["icc_profile"])

            with Image.open("Tests/images/icc_profile.png") as with_icc:
                expected_icc = with_icc.info["icc_profile"]

                im = roundtrip(im, icc_profile=expected_icc)
                self.assertEqual(im.info["icc_profile"], expected_icc)

    def test_discard_icc_profile(self):
        with Image.open("Tests/images/icc_profile.png") as im:
            im = roundtrip(im, icc_profile=None)
        self.assertNotIn("icc_profile", im.info)

    def test_roundtrip_icc_profile(self):
        with Image.open("Tests/images/icc_profile.png") as im:
            expected_icc = im.info["icc_profile"]

            im = roundtrip(im)
        self.assertEqual(im.info["icc_profile"], expected_icc)

    def test_roundtrip_no_icc_profile(self):
        with Image.open("Tests/images/icc_profile_none.png") as im:
            self.assertIsNone(im.info["icc_profile"])

            im = roundtrip(im)
        self.assertNotIn("icc_profile", im.info)

    def test_repr_png(self):
        im = hopper()

        with Image.open(BytesIO(im._repr_png_())) as repr_png:
            self.assertEqual(repr_png.format, "PNG")
            self.assert_image_equal(im, repr_png)

    def test_chunk_order(self):
        with Image.open("Tests/images/icc_profile.png") as im:
            test_file = self.tempfile("temp.png")
            im.convert("P").save(test_file, dpi=(100, 100))

        chunks = self.get_chunks(test_file)

        # https://www.w3.org/TR/PNG/#5ChunkOrdering
        # IHDR - shall be first
        self.assertEqual(chunks.index(b"IHDR"), 0)
        # PLTE - before first IDAT
        self.assertLess(chunks.index(b"PLTE"), chunks.index(b"IDAT"))
        # iCCP - before PLTE and IDAT
        self.assertLess(chunks.index(b"iCCP"), chunks.index(b"PLTE"))
        self.assertLess(chunks.index(b"iCCP"), chunks.index(b"IDAT"))
        # tRNS - after PLTE, before IDAT
        self.assertGreater(chunks.index(b"tRNS"), chunks.index(b"PLTE"))
        self.assertLess(chunks.index(b"tRNS"), chunks.index(b"IDAT"))
        # pHYs - before IDAT
        self.assertLess(chunks.index(b"pHYs"), chunks.index(b"IDAT"))

    def test_getchunks(self):
        im = hopper()

        chunks = PngImagePlugin.getchunks(im)
        self.assertEqual(len(chunks), 3)

    def test_textual_chunks_after_idat(self):
        with Image.open("Tests/images/hopper.png") as im:
            self.assertIn("comment", im.text.keys())
            for k, v in {
                "date:create": "2014-09-04T09:37:08+03:00",
                "date:modify": "2014-09-04T09:37:08+03:00",
            }.items():
                self.assertEqual(im.text[k], v)

        # Raises a SyntaxError in load_end
        with Image.open("Tests/images/broken_data_stream.png") as im:
            with self.assertRaises(IOError):
                self.assertIsInstance(im.text, dict)

        # Raises a UnicodeDecodeError in load_end
        with Image.open("Tests/images/truncated_image.png") as im:
            # The file is truncated
            self.assertRaises(IOError, lambda: im.text)
            ImageFile.LOAD_TRUNCATED_IMAGES = True
            self.assertIsInstance(im.text, dict)
            ImageFile.LOAD_TRUNCATED_IMAGES = False

        # Raises an EOFError in load_end
        with Image.open("Tests/images/hopper_idat_after_image_end.png") as im:
            self.assertEqual(im.text, {"TXT": "VALUE", "ZIP": "VALUE"})

    def test_exif(self):
        with Image.open("Tests/images/exif.png") as im:
            exif = im._getexif()
        self.assertEqual(exif[274], 1)

    def test_exif_save(self):
        with Image.open("Tests/images/exif.png") as im:
            test_file = self.tempfile("temp.png")
            im.save(test_file)

        with Image.open(test_file) as reloaded:
            exif = reloaded._getexif()
        self.assertEqual(exif[274], 1)

    def test_exif_from_jpg(self):
        with Image.open("Tests/images/pil_sample_rgb.jpg") as im:
            test_file = self.tempfile("temp.png")
            im.save(test_file)

        with Image.open(test_file) as reloaded:
            exif = reloaded._getexif()
        self.assertEqual(exif[305], "Adobe Photoshop CS Macintosh")

    def test_exif_argument(self):
        with Image.open(TEST_PNG_FILE) as im:
            test_file = self.tempfile("temp.png")
            im.save(test_file, exif=b"exifstring")

        with Image.open(test_file) as reloaded:
            self.assertEqual(reloaded.info["exif"], b"Exif\x00\x00exifstring")

    @unittest.skipUnless(
        HAVE_WEBP and _webp.HAVE_WEBPANIM, "WebP support not installed with animation"
    )
    def test_apng(self):
        with Image.open("Tests/images/iss634.apng") as im:
            self.assertEqual(im.get_format_mimetype(), "image/apng")

            # This also tests reading unknown PNG chunks (fcTL and fdAT) in load_end
            with Image.open("Tests/images/iss634.webp") as expected:
                self.assert_image_similar(im, expected, 0.23)


@unittest.skipIf(is_win32(), "requires Unix or macOS")
class TestTruncatedPngPLeaks(PillowLeakTestCase):
    mem_limit = 2 * 1024  # max increase in K
    iterations = 100  # Leak is 56k/iteration, this will leak 5.6megs

    def setUp(self):
        if "zip_encoder" not in codecs or "zip_decoder" not in codecs:
            self.skipTest("zip/deflate support not available")

    def test_leak_load(self):
        with open("Tests/images/hopper.png", "rb") as f:
            DATA = BytesIO(f.read(16 * 1024))

        ImageFile.LOAD_TRUNCATED_IMAGES = True
        with Image.open(DATA) as im:
            im.load()

        def core():
            with Image.open(DATA) as im:
                im.load()

        try:
            self._test_leak(core)
        finally:
            ImageFile.LOAD_TRUNCATED_IMAGES = False



from PIL import Image

from .helper import PillowTestCase


class TestFileBlp(PillowTestCase):
    def test_load_blp2_raw(self):
        with Image.open("Tests/images/blp/blp2_raw.blp") as im:
            with Image.open("Tests/images/blp/blp2_raw.png") as target:
                self.assert_image_equal(im, target)

    def test_load_blp2_dxt1(self):
        with Image.open("Tests/images/blp/blp2_dxt1.blp") as im:
            with Image.open("Tests/images/blp/blp2_dxt1.png") as target:
                self.assert_image_equal(im, target)

    def test_load_blp2_dxt1a(self):
        with Image.open("Tests/images/blp/blp2_dxt1a.blp") as im:
            with Image.open("Tests/images/blp/blp2_dxt1a.png") as target:
                self.assert_image_equal(im, target)



import unittest

from PIL import Image

from .helper import PillowTestCase


class TestJ2kEncodeOverflow(PillowTestCase):
    def test_j2k_overflow(self):

        im = Image.new("RGBA", (1024, 131584))
        target = self.tempfile("temp.jpc")
        with self.assertRaises(IOError):
            im.save(target)


if __name__ == "__main__":
    unittest.main()



import os
import subprocess
import sys
from unittest import TestCase


class TestMain(TestCase):
    def test_main(self):
        out = subprocess.check_output([sys.executable, "-m", "PIL"]).decode("utf-8")
        lines = out.splitlines()
        self.assertEqual(lines[0], "-" * 68)
        self.assertTrue(lines[1].startswith("Pillow "))
        self.assertTrue(lines[2].startswith("Python "))
        lines = lines[3:]
        while lines[0].startswith("    "):
            lines = lines[1:]
        self.assertEqual(lines[0], "-" * 68)
        self.assertTrue(lines[1].startswith("Python modules loaded from "))
        self.assertTrue(lines[2].startswith("Binary modules loaded from "))
        self.assertEqual(lines[3], "-" * 68)
        jpeg = (
            os.linesep
            + "-" * 68
            + os.linesep
            + "JPEG image/jpeg"
            + os.linesep
            + "Extensions: .jfif, .jpe, .jpeg, .jpg"
            + os.linesep
            + "Features: open, save"
            + os.linesep
            + "-" * 68
            + os.linesep
        )
        self.assertIn(jpeg, out)



from PIL import Hdf5StubImagePlugin, Image

from .helper import PillowTestCase

TEST_FILE = "Tests/images/hdf5.h5"


class TestFileHdf5Stub(PillowTestCase):
    def test_open(self):
        # Act
        with Image.open(TEST_FILE) as im:

            # Assert
            self.assertEqual(im.format, "HDF5")

            # Dummy data from the stub
            self.assertEqual(im.mode, "F")
            self.assertEqual(im.size, (1, 1))

    def test_invalid_file(self):
        # Arrange
        invalid_file = "Tests/images/flower.jpg"

        # Act / Assert
        self.assertRaises(
            SyntaxError, Hdf5StubImagePlugin.HDF5StubImageFile, invalid_file
        )

    def test_load(self):
        # Arrange
        with Image.open(TEST_FILE) as im:

            # Act / Assert: stub cannot load without an implemented handler
            self.assertRaises(IOError, im.load)

    def test_save(self):
        # Arrange
        with Image.open(TEST_FILE) as im:
            dummy_fp = None
            dummy_filename = "dummy.filename"

            # Act / Assert: stub cannot save without an implemented handler
            self.assertRaises(IOError, im.save, dummy_filename)
            self.assertRaises(
                IOError, Hdf5StubImagePlugin._save, im, dummy_fp, dummy_filename
            )



import unittest

from PIL import DcxImagePlugin, Image

from .helper import PillowTestCase, hopper, is_pypy

# Created with ImageMagick: convert hopper.ppm hopper.dcx
TEST_FILE = "Tests/images/hopper.dcx"


class TestFileDcx(PillowTestCase):
    def test_sanity(self):
        # Arrange

        # Act
        with Image.open(TEST_FILE) as im:

            # Assert
            self.assertEqual(im.size, (128, 128))
            self.assertIsInstance(im, DcxImagePlugin.DcxImageFile)
            orig = hopper()
            self.assert_image_equal(im, orig)

    @unittest.skipIf(is_pypy(), "Requires CPython")
    def test_unclosed_file(self):
        def open():
            im = Image.open(TEST_FILE)
            im.load()

        self.assert_warning(ResourceWarning, open)

    def test_closed_file(self):
        def open():
            im = Image.open(TEST_FILE)
            im.load()
            im.close()

        self.assert_warning(None, open)

    def test_context_manager(self):
        def open():
            with Image.open(TEST_FILE) as im:
                im.load()

        self.assert_warning(None, open)

    def test_invalid_file(self):
        with open("Tests/images/flower.jpg", "rb") as fp:
            self.assertRaises(SyntaxError, DcxImagePlugin.DcxImageFile, fp)

    def test_tell(self):
        # Arrange
        with Image.open(TEST_FILE) as im:

            # Act
            frame = im.tell()

            # Assert
            self.assertEqual(frame, 0)

    def test_n_frames(self):
        with Image.open(TEST_FILE) as im:
            self.assertEqual(im.n_frames, 1)
            self.assertFalse(im.is_animated)

    def test_eoferror(self):
        with Image.open(TEST_FILE) as im:
            n_frames = im.n_frames

            # Test seeking past the last frame
            self.assertRaises(EOFError, im.seek, n_frames)
            self.assertLess(im.tell(), n_frames)

            # Test that seeking to the last frame does not raise an error
            im.seek(n_frames - 1)

    def test_seek_too_far(self):
        # Arrange
        with Image.open(TEST_FILE) as im:
            frame = 999  # too big on purpose

        # Act / Assert
        self.assertRaises(EOFError, im.seek, frame)



from PIL import Image

from .helper import PillowTestCase

try:
    from PIL import _webp

    HAVE_WEBP = True
except ImportError:
    HAVE_WEBP = False


class TestFileWebpAnimation(PillowTestCase):
    def setUp(self):
        if not HAVE_WEBP:
            self.skipTest("WebP support not installed")
            return

        if not _webp.HAVE_WEBPANIM:
            self.skipTest(
                "WebP library does not contain animation support, "
                "not testing animation"
            )

    def test_n_frames(self):
        """
        Ensure that WebP format sets n_frames and is_animated
        attributes correctly.
        """

        with Image.open("Tests/images/hopper.webp") as im:
            self.assertEqual(im.n_frames, 1)
            self.assertFalse(im.is_animated)

        with Image.open("Tests/images/iss634.webp") as im:
            self.assertEqual(im.n_frames, 42)
            self.assertTrue(im.is_animated)

    def test_write_animation_L(self):
        """
        Convert an animated GIF to animated WebP, then compare the
        frame count, and first and last frames to ensure they're
        visually similar.
        """

        with Image.open("Tests/images/iss634.gif") as orig:
            self.assertGreater(orig.n_frames, 1)

            temp_file = self.tempfile("temp.webp")
            orig.save(temp_file, save_all=True)
            with Image.open(temp_file) as im:
                self.assertEqual(im.n_frames, orig.n_frames)

                # Compare first and last frames to the original animated GIF
                orig.load()
                im.load()
                self.assert_image_similar(im, orig.convert("RGBA"), 25.0)
                orig.seek(orig.n_frames - 1)
                im.seek(im.n_frames - 1)
                orig.load()
                im.load()
                self.assert_image_similar(im, orig.convert("RGBA"), 25.0)

    def test_write_animation_RGB(self):
        """
        Write an animated WebP from RGB frames, and ensure the frames
        are visually similar to the originals.
        """

        def check(temp_file):
            with Image.open(temp_file) as im:
                self.assertEqual(im.n_frames, 2)

                # Compare first frame to original
                im.load()
                self.assert_image_equal(im, frame1.convert("RGBA"))

                # Compare second frame to original
                im.seek(1)
                im.load()
                self.assert_image_equal(im, frame2.convert("RGBA"))

        with Image.open("Tests/images/anim_frame1.webp") as frame1:
            with Image.open("Tests/images/anim_frame2.webp") as frame2:
                temp_file1 = self.tempfile("temp.webp")
                frame1.copy().save(
                    temp_file1, save_all=True, append_images=[frame2], lossless=True
                )
                check(temp_file1)

                # Tests appending using a generator
                def imGenerator(ims):
                    yield from ims

                temp_file2 = self.tempfile("temp_generator.webp")
                frame1.copy().save(
                    temp_file2,
                    save_all=True,
                    append_images=imGenerator([frame2]),
                    lossless=True,
                )
                check(temp_file2)

    def test_timestamp_and_duration(self):
        """
        Try passing a list of durations, and make sure the encoded
        timestamps and durations are correct.
        """

        durations = [0, 10, 20, 30, 40]
        temp_file = self.tempfile("temp.webp")
        with Image.open("Tests/images/anim_frame1.webp") as frame1:
            with Image.open("Tests/images/anim_frame2.webp") as frame2:
                frame1.save(
                    temp_file,
                    save_all=True,
                    append_images=[frame2, frame1, frame2, frame1],
                    duration=durations,
                )

        with Image.open(temp_file) as im:
            self.assertEqual(im.n_frames, 5)
            self.assertTrue(im.is_animated)

            # Check that timestamps and durations match original values specified
            ts = 0
            for frame in range(im.n_frames):
                im.seek(frame)
                im.load()
                self.assertEqual(im.info["duration"], durations[frame])
                self.assertEqual(im.info["timestamp"], ts)
                ts += durations[frame]

    def test_seeking(self):
        """
        Create an animated WebP file, and then try seeking through
        frames in reverse-order, verifying the timestamps and durations
        are correct.
        """

        dur = 33
        temp_file = self.tempfile("temp.webp")
        with Image.open("Tests/images/anim_frame1.webp") as frame1:
            with Image.open("Tests/images/anim_frame2.webp") as frame2:
                frame1.save(
                    temp_file,
                    save_all=True,
                    append_images=[frame2, frame1, frame2, frame1],
                    duration=dur,
                )

        with Image.open(temp_file) as im:
            self.assertEqual(im.n_frames, 5)
            self.assertTrue(im.is_animated)

            # Traverse frames in reverse, checking timestamps and durations
            ts = dur * (im.n_frames - 1)
            for frame in reversed(range(im.n_frames)):
                im.seek(frame)
                im.load()
                self.assertEqual(im.info["duration"], dur)
                self.assertEqual(im.info["timestamp"], ts)
                ts -= dur



from PIL import Image, XpmImagePlugin

from .helper import PillowTestCase, hopper

TEST_FILE = "Tests/images/hopper.xpm"


class TestFileXpm(PillowTestCase):
    def test_sanity(self):
        with Image.open(TEST_FILE) as im:
            im.load()
            self.assertEqual(im.mode, "P")
            self.assertEqual(im.size, (128, 128))
            self.assertEqual(im.format, "XPM")

            # large error due to quantization->44 colors.
            self.assert_image_similar(im.convert("RGB"), hopper("RGB"), 60)

    def test_invalid_file(self):
        invalid_file = "Tests/images/flower.jpg"

        self.assertRaises(SyntaxError, XpmImagePlugin.XpmImageFile, invalid_file)

    def test_load_read(self):
        # Arrange
        with Image.open(TEST_FILE) as im:
            dummy_bytes = 1

            # Act
            data = im.load_read(dummy_bytes)

        # Assert
        self.assertEqual(len(data), 16384)



import os

from PIL import Image

from .helper import PillowTestCase, hopper


class TestImageLoad(PillowTestCase):
    def test_sanity(self):

        im = hopper()

        pix = im.load()

        self.assertEqual(pix[0, 0], (20, 20, 70))

    def test_close(self):
        im = Image.open("Tests/images/hopper.gif")
        im.close()
        self.assertRaises(ValueError, im.load)
        self.assertRaises(ValueError, im.getpixel, (0, 0))

    def test_contextmanager(self):
        fn = None
        with Image.open("Tests/images/hopper.gif") as im:
            fn = im.fp.fileno()
            os.fstat(fn)

        self.assertRaises(OSError, os.fstat, fn)

    def test_contextmanager_non_exclusive_fp(self):
        with open("Tests/images/hopper.gif", "rb") as fp:
            with Image.open(fp):
                pass

            self.assertFalse(fp.closed)



import unittest

from PIL import Image, WebPImagePlugin

from .helper import PillowTestCase, hopper

try:
    from PIL import _webp

    HAVE_WEBP = True
except ImportError:
    HAVE_WEBP = False


class TestUnsupportedWebp(PillowTestCase):
    def test_unsupported(self):
        if HAVE_WEBP:
            WebPImagePlugin.SUPPORTED = False

        file_path = "Tests/images/hopper.webp"
        self.assert_warning(
            UserWarning, lambda: self.assertRaises(IOError, Image.open, file_path)
        )

        if HAVE_WEBP:
            WebPImagePlugin.SUPPORTED = True


@unittest.skipUnless(HAVE_WEBP, "WebP support not installed")
class TestFileWebp(PillowTestCase):
    def setUp(self):
        self.rgb_mode = "RGB"

    def test_version(self):
        _webp.WebPDecoderVersion()
        _webp.WebPDecoderBuggyAlpha()

    def test_read_rgb(self):
        """
        Can we read a RGB mode WebP file without error?
        Does it have the bits we expect?
        """

        with Image.open("Tests/images/hopper.webp") as image:
            self.assertEqual(image.mode, self.rgb_mode)
            self.assertEqual(image.size, (128, 128))
            self.assertEqual(image.format, "WEBP")
            image.load()
            image.getdata()

            # generated with:
            # dwebp -ppm ../../Tests/images/hopper.webp -o hopper_webp_bits.ppm
            self.assert_image_similar_tofile(
                image, "Tests/images/hopper_webp_bits.ppm", 1.0
            )

    def test_write_rgb(self):
        """
        Can we write a RGB mode file to webp without error.
        Does it have the bits we expect?
        """

        temp_file = self.tempfile("temp.webp")

        hopper(self.rgb_mode).save(temp_file)
        with Image.open(temp_file) as image:
            self.assertEqual(image.mode, self.rgb_mode)
            self.assertEqual(image.size, (128, 128))
            self.assertEqual(image.format, "WEBP")
            image.load()
            image.getdata()

            # generated with: dwebp -ppm temp.webp -o hopper_webp_write.ppm
            self.assert_image_similar_tofile(
                image, "Tests/images/hopper_webp_write.ppm", 12.0
            )

            # This test asserts that the images are similar. If the average pixel
            # difference between the two images is less than the epsilon value,
            # then we're going to accept that it's a reasonable lossy version of
            # the image. The old lena images for WebP are showing ~16 on
            # Ubuntu, the jpegs are showing ~18.
            target = hopper(self.rgb_mode)
            self.assert_image_similar(image, target, 12.0)

    def test_write_unsupported_mode_L(self):
        """
        Saving a black-and-white file to WebP format should work, and be
        similar to the original file.
        """

        temp_file = self.tempfile("temp.webp")
        hopper("L").save(temp_file)
        with Image.open(temp_file) as image:
            self.assertEqual(image.mode, self.rgb_mode)
            self.assertEqual(image.size, (128, 128))
            self.assertEqual(image.format, "WEBP")

            image.load()
            image.getdata()
            target = hopper("L").convert(self.rgb_mode)

            self.assert_image_similar(image, target, 10.0)

    def test_write_unsupported_mode_P(self):
        """
        Saving a palette-based file to WebP format should work, and be
        similar to the original file.
        """

        temp_file = self.tempfile("temp.webp")
        hopper("P").save(temp_file)
        with Image.open(temp_file) as image:
            self.assertEqual(image.mode, self.rgb_mode)
            self.assertEqual(image.size, (128, 128))
            self.assertEqual(image.format, "WEBP")

            image.load()
            image.getdata()
            target = hopper("P").convert(self.rgb_mode)

            self.assert_image_similar(image, target, 50.0)

    def test_WebPEncode_with_invalid_args(self):
        """
        Calling encoder functions with no arguments should result in an error.
        """

        if _webp.HAVE_WEBPANIM:
            self.assertRaises(TypeError, _webp.WebPAnimEncoder)
        self.assertRaises(TypeError, _webp.WebPEncode)

    def test_WebPDecode_with_invalid_args(self):
        """
        Calling decoder functions with no arguments should result in an error.
        """

        if _webp.HAVE_WEBPANIM:
            self.assertRaises(TypeError, _webp.WebPAnimDecoder)
        self.assertRaises(TypeError, _webp.WebPDecode)

    def test_no_resource_warning(self):
        file_path = "Tests/images/hopper.webp"
        with Image.open(file_path) as image:
            temp_file = self.tempfile("temp.webp")
            self.assert_warning(None, image.save, temp_file)

    def test_file_pointer_could_be_reused(self):
        file_path = "Tests/images/hopper.webp"
        with open(file_path, "rb") as blob:
            Image.open(blob).load()
            Image.open(blob).load()

    @unittest.skipUnless(
        HAVE_WEBP and _webp.HAVE_WEBPANIM, "WebP save all not available"
    )
    def test_background_from_gif(self):
        with Image.open("Tests/images/chi.gif") as im:
            original_value = im.convert("RGB").getpixel((1, 1))

            # Save as WEBP
            out_webp = self.tempfile("temp.webp")
            im.save(out_webp, save_all=True)

        # Save as GIF
        out_gif = self.tempfile("temp.gif")
        Image.open(out_webp).save(out_gif)

        with Image.open(out_gif) as reread:
            reread_value = reread.convert("RGB").getpixel((1, 1))
        difference = sum(
            [abs(original_value[i] - reread_value[i]) for i in range(0, 3)]
        )
        self.assertLess(difference, 5)



from PIL import Image

from .helper import PillowTestCase, hopper


class TestImageRotate(PillowTestCase):
    def rotate(self, im, mode, angle, center=None, translate=None):
        out = im.rotate(angle, center=center, translate=translate)
        self.assertEqual(out.mode, mode)
        self.assertEqual(out.size, im.size)  # default rotate clips output
        out = im.rotate(angle, center=center, translate=translate, expand=1)
        self.assertEqual(out.mode, mode)
        if angle % 180 == 0:
            self.assertEqual(out.size, im.size)
        elif im.size == (0, 0):
            self.assertEqual(out.size, im.size)
        else:
            self.assertNotEqual(out.size, im.size)

    def test_mode(self):
        for mode in ("1", "P", "L", "RGB", "I", "F"):
            im = hopper(mode)
            self.rotate(im, mode, 45)

    def test_angle(self):
        for angle in (0, 90, 180, 270):
            with Image.open("Tests/images/test-card.png") as im:
                self.rotate(im, im.mode, angle)

    def test_zero(self):
        for angle in (0, 45, 90, 180, 270):
            im = Image.new("RGB", (0, 0))
            self.rotate(im, im.mode, angle)

    def test_resample(self):
        # Target image creation, inspected by eye.
        # >>> im = Image.open('Tests/images/hopper.ppm')
        # >>> im = im.rotate(45, resample=Image.BICUBIC, expand=True)
        # >>> im.save('Tests/images/hopper_45.png')

        with Image.open("Tests/images/hopper_45.png") as target:
            for (resample, epsilon) in (
                (Image.NEAREST, 10),
                (Image.BILINEAR, 5),
                (Image.BICUBIC, 0),
            ):
                im = hopper()
                im = im.rotate(45, resample=resample, expand=True)
                self.assert_image_similar(im, target, epsilon)

    def test_center_0(self):
        im = hopper()
        im = im.rotate(45, center=(0, 0), resample=Image.BICUBIC)

        with Image.open("Tests/images/hopper_45.png") as target:
            target_origin = target.size[1] / 2
            target = target.crop((0, target_origin, 128, target_origin + 128))

        self.assert_image_similar(im, target, 15)

    def test_center_14(self):
        im = hopper()
        im = im.rotate(45, center=(14, 14), resample=Image.BICUBIC)

        with Image.open("Tests/images/hopper_45.png") as target:
            target_origin = target.size[1] / 2 - 14
            target = target.crop((6, target_origin, 128 + 6, target_origin + 128))

            self.assert_image_similar(im, target, 10)

    def test_translate(self):
        im = hopper()
        with Image.open("Tests/images/hopper_45.png") as target:
            target_origin = (target.size[1] / 2 - 64) - 5
            target = target.crop(
                (target_origin, target_origin, target_origin + 128, target_origin + 128)
            )

        im = im.rotate(45, translate=(5, 5), resample=Image.BICUBIC)

        self.assert_image_similar(im, target, 1)

    def test_fastpath_center(self):
        # if the center is -1,-1 and we rotate by 90<=x<=270 the
        # resulting image should be black
        for angle in (90, 180, 270):
            im = hopper().rotate(angle, center=(-1, -1))
            self.assert_image_equal(im, Image.new("RGB", im.size, "black"))

    def test_fastpath_translate(self):
        # if we post-translate by -128
        # resulting image should be black
        for angle in (0, 90, 180, 270):
            im = hopper().rotate(angle, translate=(-128, -128))
            self.assert_image_equal(im, Image.new("RGB", im.size, "black"))

    def test_center(self):
        im = hopper()
        self.rotate(im, im.mode, 45, center=(0, 0))
        self.rotate(im, im.mode, 45, translate=(im.size[0] / 2, 0))
        self.rotate(im, im.mode, 45, center=(0, 0), translate=(im.size[0] / 2, 0))

    def test_rotate_no_fill(self):
        im = Image.new("RGB", (100, 100), "green")
        im = im.rotate(45)
        with Image.open("Tests/images/rotate_45_no_fill.png") as target:
            self.assert_image_equal(im, target)

    def test_rotate_with_fill(self):
        im = Image.new("RGB", (100, 100), "green")
        im = im.rotate(45, fillcolor="white")
        with Image.open("Tests/images/rotate_45_with_fill.png") as target:
            self.assert_image_equal(im, target)

    def test_alpha_rotate_no_fill(self):
        # Alpha images are handled differently internally
        im = Image.new("RGBA", (10, 10), "green")
        im = im.rotate(45, expand=1)
        corner = im.getpixel((0, 0))
        self.assertEqual(corner, (0, 0, 0, 0))

    def test_alpha_rotate_with_fill(self):
        # Alpha images are handled differently internally
        im = Image.new("RGBA", (10, 10), "green")
        im = im.rotate(45, expand=1, fillcolor=(255, 0, 0, 255))
        corner = im.getpixel((0, 0))
        self.assertEqual(corner, (255, 0, 0, 255))



import unittest

from PIL import FliImagePlugin, Image

from .helper import PillowTestCase, is_pypy

# created as an export of a palette image from Gimp2.6
# save as...-> hopper.fli, default options.
static_test_file = "Tests/images/hopper.fli"

# From https://samples.libav.org/fli-flc/
animated_test_file = "Tests/images/a.fli"


class TestFileFli(PillowTestCase):
    def test_sanity(self):
        with Image.open(static_test_file) as im:
            im.load()
            self.assertEqual(im.mode, "P")
            self.assertEqual(im.size, (128, 128))
            self.assertEqual(im.format, "FLI")
            self.assertFalse(im.is_animated)

        with Image.open(animated_test_file) as im:
            self.assertEqual(im.mode, "P")
            self.assertEqual(im.size, (320, 200))
            self.assertEqual(im.format, "FLI")
            self.assertEqual(im.info["duration"], 71)
            self.assertTrue(im.is_animated)

    @unittest.skipIf(is_pypy(), "Requires CPython")
    def test_unclosed_file(self):
        def open():
            im = Image.open(static_test_file)
            im.load()

        self.assert_warning(ResourceWarning, open)

    def test_closed_file(self):
        def open():
            im = Image.open(static_test_file)
            im.load()
            im.close()

        self.assert_warning(None, open)

    def test_context_manager(self):
        def open():
            with Image.open(static_test_file) as im:
                im.load()

        self.assert_warning(None, open)

    def test_tell(self):
        # Arrange
        with Image.open(static_test_file) as im:

            # Act
            frame = im.tell()

            # Assert
            self.assertEqual(frame, 0)

    def test_invalid_file(self):
        invalid_file = "Tests/images/flower.jpg"

        self.assertRaises(SyntaxError, FliImagePlugin.FliImageFile, invalid_file)

    def test_n_frames(self):
        with Image.open(static_test_file) as im:
            self.assertEqual(im.n_frames, 1)
            self.assertFalse(im.is_animated)

        with Image.open(animated_test_file) as im:
            self.assertEqual(im.n_frames, 384)
            self.assertTrue(im.is_animated)

    def test_eoferror(self):
        with Image.open(animated_test_file) as im:
            n_frames = im.n_frames

            # Test seeking past the last frame
            self.assertRaises(EOFError, im.seek, n_frames)
            self.assertLess(im.tell(), n_frames)

            # Test that seeking to the last frame does not raise an error
            im.seek(n_frames - 1)

    def test_seek_tell(self):
        with Image.open(animated_test_file) as im:

            layer_number = im.tell()
            self.assertEqual(layer_number, 0)

            im.seek(0)
            layer_number = im.tell()
            self.assertEqual(layer_number, 0)

            im.seek(1)
            layer_number = im.tell()
            self.assertEqual(layer_number, 1)

            im.seek(2)
            layer_number = im.tell()
            self.assertEqual(layer_number, 2)

            im.seek(1)
            layer_number = im.tell()
            self.assertEqual(layer_number, 1)

    def test_seek(self):
        with Image.open(animated_test_file) as im:
            im.seek(50)

            with Image.open("Tests/images/a_fli.png") as expected:
                self.assert_image_equal(im, expected)



import unittest
from io import BytesIO

from PIL import GifImagePlugin, Image, ImageDraw, ImagePalette

from .helper import PillowTestCase, hopper, is_pypy, netpbm_available

try:
    from PIL import _webp

    HAVE_WEBP = True
except ImportError:
    HAVE_WEBP = False

codecs = dir(Image.core)

# sample gif stream
TEST_GIF = "Tests/images/hopper.gif"

with open(TEST_GIF, "rb") as f:
    data = f.read()


class TestFileGif(PillowTestCase):
    def setUp(self):
        if "gif_encoder" not in codecs or "gif_decoder" not in codecs:
            self.skipTest("gif support not available")  # can this happen?

    def test_sanity(self):
        with Image.open(TEST_GIF) as im:
            im.load()
            self.assertEqual(im.mode, "P")
            self.assertEqual(im.size, (128, 128))
            self.assertEqual(im.format, "GIF")
            self.assertEqual(im.info["version"], b"GIF89a")

    @unittest.skipIf(is_pypy(), "Requires CPython")
    def test_unclosed_file(self):
        def open():
            im = Image.open(TEST_GIF)
            im.load()

        self.assert_warning(ResourceWarning, open)

    def test_closed_file(self):
        def open():
            im = Image.open(TEST_GIF)
            im.load()
            im.close()

        self.assert_warning(None, open)

    def test_context_manager(self):
        def open():
            with Image.open(TEST_GIF) as im:
                im.load()

        self.assert_warning(None, open)

    def test_invalid_file(self):
        invalid_file = "Tests/images/flower.jpg"

        self.assertRaises(SyntaxError, GifImagePlugin.GifImageFile, invalid_file)

    def test_optimize(self):
        def test_grayscale(optimize):
            im = Image.new("L", (1, 1), 0)
            filename = BytesIO()
            im.save(filename, "GIF", optimize=optimize)
            return len(filename.getvalue())

        def test_bilevel(optimize):
            im = Image.new("1", (1, 1), 0)
            test_file = BytesIO()
            im.save(test_file, "GIF", optimize=optimize)
            return len(test_file.getvalue())

        self.assertEqual(test_grayscale(0), 800)
        self.assertEqual(test_grayscale(1), 44)
        self.assertEqual(test_bilevel(0), 800)
        self.assertEqual(test_bilevel(1), 800)

    def test_optimize_correctness(self):
        # 256 color Palette image, posterize to > 128 and < 128 levels
        # Size bigger and smaller than 512x512
        # Check the palette for number of colors allocated.
        # Check for correctness after conversion back to RGB
        def check(colors, size, expected_palette_length):
            # make an image with empty colors in the start of the palette range
            im = Image.frombytes(
                "P", (colors, colors), bytes(range(256 - colors, 256)) * colors
            )
            im = im.resize((size, size))
            outfile = BytesIO()
            im.save(outfile, "GIF")
            outfile.seek(0)
            with Image.open(outfile) as reloaded:
                # check palette length
                palette_length = max(
                    i + 1 for i, v in enumerate(reloaded.histogram()) if v
                )
                self.assertEqual(expected_palette_length, palette_length)

                self.assert_image_equal(im.convert("RGB"), reloaded.convert("RGB"))

        # These do optimize the palette
        check(128, 511, 128)
        check(64, 511, 64)
        check(4, 511, 4)

        # These don't optimize the palette
        check(128, 513, 256)
        check(64, 513, 256)
        check(4, 513, 256)

        # other limits that don't optimize the palette
        check(129, 511, 256)
        check(255, 511, 256)
        check(256, 511, 256)

    def test_optimize_full_l(self):
        im = Image.frombytes("L", (16, 16), bytes(range(256)))
        test_file = BytesIO()
        im.save(test_file, "GIF", optimize=True)
        self.assertEqual(im.mode, "L")

    def test_roundtrip(self):
        out = self.tempfile("temp.gif")
        im = hopper()
        im.save(out)
        with Image.open(out) as reread:

            self.assert_image_similar(reread.convert("RGB"), im, 50)

    def test_roundtrip2(self):
        # see https://github.com/python-pillow/Pillow/issues/403
        out = self.tempfile("temp.gif")
        with Image.open(TEST_GIF) as im:
            im2 = im.copy()
            im2.save(out)
        with Image.open(out) as reread:

            self.assert_image_similar(reread.convert("RGB"), hopper(), 50)

    def test_roundtrip_save_all(self):
        # Single frame image
        out = self.tempfile("temp.gif")
        im = hopper()
        im.save(out, save_all=True)
        with Image.open(out) as reread:

            self.assert_image_similar(reread.convert("RGB"), im, 50)

        # Multiframe image
        with Image.open("Tests/images/dispose_bgnd.gif") as im:

            out = self.tempfile("temp.gif")
            im.save(out, save_all=True)
        with Image.open(out) as reread:

            self.assertEqual(reread.n_frames, 5)

    def test_headers_saving_for_animated_gifs(self):
        important_headers = ["background", "version", "duration", "loop"]
        # Multiframe image
        with Image.open("Tests/images/dispose_bgnd.gif") as im:

            info = im.info.copy()

            out = self.tempfile("temp.gif")
            im.save(out, save_all=True)
        with Image.open(out) as reread:

            for header in important_headers:
                self.assertEqual(info[header], reread.info[header])

    def test_palette_handling(self):
        # see https://github.com/python-pillow/Pillow/issues/513

        with Image.open(TEST_GIF) as im:
            im = im.convert("RGB")

            im = im.resize((100, 100), Image.LANCZOS)
            im2 = im.convert("P", palette=Image.ADAPTIVE, colors=256)

            f = self.tempfile("temp.gif")
            im2.save(f, optimize=True)

        with Image.open(f) as reloaded:

            self.assert_image_similar(im, reloaded.convert("RGB"), 10)

    def test_palette_434(self):
        # see https://github.com/python-pillow/Pillow/issues/434

        def roundtrip(im, *args, **kwargs):
            out = self.tempfile("temp.gif")
            im.copy().save(out, *args, **kwargs)
            reloaded = Image.open(out)

            return reloaded

        orig = "Tests/images/test.colors.gif"
        with Image.open(orig) as im:

            with roundtrip(im) as reloaded:
                self.assert_image_similar(im, reloaded, 1)
            with roundtrip(im, optimize=True) as reloaded:
                self.assert_image_similar(im, reloaded, 1)

            im = im.convert("RGB")
            # check automatic P conversion
            with roundtrip(im) as reloaded:
                reloaded = reloaded.convert("RGB")
                self.assert_image_equal(im, reloaded)

    @unittest.skipUnless(netpbm_available(), "netpbm not available")
    def test_save_netpbm_bmp_mode(self):
        with Image.open(TEST_GIF) as img:
            img = img.convert("RGB")

            tempfile = self.tempfile("temp.gif")
            GifImagePlugin._save_netpbm(img, 0, tempfile)
            with Image.open(tempfile) as reloaded:
                self.assert_image_similar(img, reloaded.convert("RGB"), 0)

    @unittest.skipUnless(netpbm_available(), "netpbm not available")
    def test_save_netpbm_l_mode(self):
        with Image.open(TEST_GIF) as img:
            img = img.convert("L")

            tempfile = self.tempfile("temp.gif")
            GifImagePlugin._save_netpbm(img, 0, tempfile)
            with Image.open(tempfile) as reloaded:
                self.assert_image_similar(img, reloaded.convert("L"), 0)

    def test_seek(self):
        with Image.open("Tests/images/dispose_none.gif") as img:
            framecount = 0
            try:
                while True:
                    framecount += 1
                    img.seek(img.tell() + 1)
            except EOFError:
                self.assertEqual(framecount, 5)

    def test_seek_info(self):
        with Image.open("Tests/images/iss634.gif") as im:
            info = im.info.copy()

            im.seek(1)
            im.seek(0)

            self.assertEqual(im.info, info)

    def test_seek_rewind(self):
        with Image.open("Tests/images/iss634.gif") as im:
            im.seek(2)
            im.seek(1)

            with Image.open("Tests/images/iss634.gif") as expected:
                expected.seek(1)
                self.assert_image_equal(im, expected)

    def test_n_frames(self):
        for path, n_frames in [[TEST_GIF, 1], ["Tests/images/iss634.gif", 42]]:
            # Test is_animated before n_frames
            with Image.open(path) as im:
                self.assertEqual(im.is_animated, n_frames != 1)

            # Test is_animated after n_frames
            with Image.open(path) as im:
                self.assertEqual(im.n_frames, n_frames)
                self.assertEqual(im.is_animated, n_frames != 1)

    def test_eoferror(self):
        with Image.open(TEST_GIF) as im:
            n_frames = im.n_frames

            # Test seeking past the last frame
            self.assertRaises(EOFError, im.seek, n_frames)
            self.assertLess(im.tell(), n_frames)

            # Test that seeking to the last frame does not raise an error
            im.seek(n_frames - 1)

    def test_dispose_none(self):
        with Image.open("Tests/images/dispose_none.gif") as img:
            try:
                while True:
                    img.seek(img.tell() + 1)
                    self.assertEqual(img.disposal_method, 1)
            except EOFError:
                pass

    def test_dispose_background(self):
        with Image.open("Tests/images/dispose_bgnd.gif") as img:
            try:
                while True:
                    img.seek(img.tell() + 1)
                    self.assertEqual(img.disposal_method, 2)
            except EOFError:
                pass

    def test_dispose_previous(self):
        with Image.open("Tests/images/dispose_prev.gif") as img:
            try:
                while True:
                    img.seek(img.tell() + 1)
                    self.assertEqual(img.disposal_method, 3)
            except EOFError:
                pass

    def test_save_dispose(self):
        out = self.tempfile("temp.gif")
        im_list = [
            Image.new("L", (100, 100), "#000"),
            Image.new("L", (100, 100), "#111"),
            Image.new("L", (100, 100), "#222"),
        ]
        for method in range(0, 4):
            im_list[0].save(
                out, save_all=True, append_images=im_list[1:], disposal=method
            )
            with Image.open(out) as img:
                for _ in range(2):
                    img.seek(img.tell() + 1)
                    self.assertEqual(img.disposal_method, method)

        # check per frame disposal
        im_list[0].save(
            out,
            save_all=True,
            append_images=im_list[1:],
            disposal=tuple(range(len(im_list))),
        )

        with Image.open(out) as img:

            for i in range(2):
                img.seek(img.tell() + 1)
                self.assertEqual(img.disposal_method, i + 1)

    def test_dispose2_palette(self):
        out = self.tempfile("temp.gif")

        # 4 backgrounds: White, Grey, Black, Red
        circles = [(255, 255, 255), (153, 153, 153), (0, 0, 0), (255, 0, 0)]

        im_list = []
        for circle in circles:
            img = Image.new("RGB", (100, 100), (255, 0, 0))

            # Red circle in center of each frame
            d = ImageDraw.Draw(img)
            d.ellipse([(40, 40), (60, 60)], fill=circle)

            im_list.append(img)

        im_list[0].save(out, save_all=True, append_images=im_list[1:], disposal=2)

        with Image.open(out) as img:
            for i, circle in enumerate(circles):
                img.seek(i)
                rgb_img = img.convert("RGB")

                # Check top left pixel matches background
                self.assertEqual(rgb_img.getpixel((0, 0)), (255, 0, 0))

                # Center remains red every frame
                self.assertEqual(rgb_img.getpixel((50, 50)), circle)

    def test_dispose2_diff(self):
        out = self.tempfile("temp.gif")

        # 4 frames: red/blue, red/red, blue/blue, red/blue
        circles = [
            ((255, 0, 0, 255), (0, 0, 255, 255)),
            ((255, 0, 0, 255), (255, 0, 0, 255)),
            ((0, 0, 255, 255), (0, 0, 255, 255)),
            ((255, 0, 0, 255), (0, 0, 255, 255)),
        ]

        im_list = []
        for i in range(len(circles)):
            # Transparent BG
            img = Image.new("RGBA", (100, 100), (255, 255, 255, 0))

            # Two circles per frame
            d = ImageDraw.Draw(img)
            d.ellipse([(0, 30), (40, 70)], fill=circles[i][0])
            d.ellipse([(60, 30), (100, 70)], fill=circles[i][1])

            im_list.append(img)

        im_list[0].save(
            out, save_all=True, append_images=im_list[1:], disposal=2, transparency=0
        )

        with Image.open(out) as img:
            for i, colours in enumerate(circles):
                img.seek(i)
                rgb_img = img.convert("RGBA")

                # Check left circle is correct colour
                self.assertEqual(rgb_img.getpixel((20, 50)), colours[0])

                # Check right circle is correct colour
                self.assertEqual(rgb_img.getpixel((80, 50)), colours[1])

                # Check BG is correct colour
                self.assertEqual(rgb_img.getpixel((1, 1)), (255, 255, 255, 0))

    def test_dispose2_background(self):
        out = self.tempfile("temp.gif")

        im_list = []

        im = Image.new("P", (100, 100))
        d = ImageDraw.Draw(im)
        d.rectangle([(50, 0), (100, 100)], fill="#f00")
        d.rectangle([(0, 0), (50, 100)], fill="#0f0")
        im_list.append(im)

        im = Image.new("P", (100, 100))
        d = ImageDraw.Draw(im)
        d.rectangle([(0, 0), (100, 50)], fill="#f00")
        d.rectangle([(0, 50), (100, 100)], fill="#0f0")
        im_list.append(im)

        im_list[0].save(
            out, save_all=True, append_images=im_list[1:], disposal=[0, 2], background=1
        )

        with Image.open(out) as im:
            im.seek(1)
            self.assertEqual(im.getpixel((0, 0)), 0)

    def test_iss634(self):
        with Image.open("Tests/images/iss634.gif") as img:
            # seek to the second frame
            img.seek(img.tell() + 1)
            # all transparent pixels should be replaced with the color from the
            # first frame
            self.assertEqual(img.histogram()[img.info["transparency"]], 0)

    def test_duration(self):
        duration = 1000

        out = self.tempfile("temp.gif")
        im = Image.new("L", (100, 100), "#000")

        # Check that the argument has priority over the info settings
        im.info["duration"] = 100
        im.save(out, duration=duration)

        with Image.open(out) as reread:
            self.assertEqual(reread.info["duration"], duration)

    def test_multiple_duration(self):
        duration_list = [1000, 2000, 3000]

        out = self.tempfile("temp.gif")
        im_list = [
            Image.new("L", (100, 100), "#000"),
            Image.new("L", (100, 100), "#111"),
            Image.new("L", (100, 100), "#222"),
        ]

        # duration as list
        im_list[0].save(
            out, save_all=True, append_images=im_list[1:], duration=duration_list
        )
        with Image.open(out) as reread:

            for duration in duration_list:
                self.assertEqual(reread.info["duration"], duration)
                try:
                    reread.seek(reread.tell() + 1)
                except EOFError:
                    pass

        # duration as tuple
        im_list[0].save(
            out, save_all=True, append_images=im_list[1:], duration=tuple(duration_list)
        )
        with Image.open(out) as reread:

            for duration in duration_list:
                self.assertEqual(reread.info["duration"], duration)
                try:
                    reread.seek(reread.tell() + 1)
                except EOFError:
                    pass

    def test_identical_frames(self):
        duration_list = [1000, 1500, 2000, 4000]

        out = self.tempfile("temp.gif")
        im_list = [
            Image.new("L", (100, 100), "#000"),
            Image.new("L", (100, 100), "#000"),
            Image.new("L", (100, 100), "#000"),
            Image.new("L", (100, 100), "#111"),
        ]

        # duration as list
        im_list[0].save(
            out, save_all=True, append_images=im_list[1:], duration=duration_list
        )
        with Image.open(out) as reread:

            # Assert that the first three frames were combined
            self.assertEqual(reread.n_frames, 2)

            # Assert that the new duration is the total of the identical frames
            self.assertEqual(reread.info["duration"], 4500)

    def test_identical_frames_to_single_frame(self):
        for duration in ([1000, 1500, 2000, 4000], (1000, 1500, 2000, 4000), 8500):
            out = self.tempfile("temp.gif")
            im_list = [
                Image.new("L", (100, 100), "#000"),
                Image.new("L", (100, 100), "#000"),
                Image.new("L", (100, 100), "#000"),
            ]

            im_list[0].save(
                out, save_all=True, append_images=im_list[1:], duration=duration
            )
            with Image.open(out) as reread:
                # Assert that all frames were combined
                self.assertEqual(reread.n_frames, 1)

                # Assert that the new duration is the total of the identical frames
                self.assertEqual(reread.info["duration"], 8500)

    def test_number_of_loops(self):
        number_of_loops = 2

        out = self.tempfile("temp.gif")
        im = Image.new("L", (100, 100), "#000")
        im.save(out, loop=number_of_loops)
        with Image.open(out) as reread:

            self.assertEqual(reread.info["loop"], number_of_loops)

    def test_background(self):
        out = self.tempfile("temp.gif")
        im = Image.new("L", (100, 100), "#000")
        im.info["background"] = 1
        im.save(out)
        with Image.open(out) as reread:

            self.assertEqual(reread.info["background"], im.info["background"])

        if HAVE_WEBP and _webp.HAVE_WEBPANIM:
            with Image.open("Tests/images/hopper.webp") as im:
                self.assertIsInstance(im.info["background"], tuple)
                im.save(out)

    def test_comment(self):
        with Image.open(TEST_GIF) as im:
            self.assertEqual(
                im.info["comment"], b"File written by Adobe Photoshop\xa8 4.0"
            )

        out = self.tempfile("temp.gif")
        im = Image.new("L", (100, 100), "#000")
        im.info["comment"] = b"Test comment text"
        im.save(out)
        with Image.open(out) as reread:
            self.assertEqual(reread.info["comment"], im.info["comment"])

        im.info["comment"] = "Test comment text"
        im.save(out)
        with Image.open(out) as reread:
            self.assertEqual(reread.info["comment"], im.info["comment"].encode())

    def test_comment_over_255(self):
        out = self.tempfile("temp.gif")
        im = Image.new("L", (100, 100), "#000")
        comment = b"Test comment text"
        while len(comment) < 256:
            comment += comment
        im.info["comment"] = comment
        im.save(out)
        with Image.open(out) as reread:

            self.assertEqual(reread.info["comment"], comment)

    def test_zero_comment_subblocks(self):
        with Image.open("Tests/images/hopper_zero_comment_subblocks.gif") as im:
            with Image.open(TEST_GIF) as expected:
                self.assert_image_equal(im, expected)

    def test_version(self):
        out = self.tempfile("temp.gif")

        def assertVersionAfterSave(im, version):
            im.save(out)
            with Image.open(out) as reread:
                self.assertEqual(reread.info["version"], version)

        # Test that GIF87a is used by default
        im = Image.new("L", (100, 100), "#000")
        assertVersionAfterSave(im, b"GIF87a")

        # Test setting the version to 89a
        im = Image.new("L", (100, 100), "#000")
        im.info["version"] = b"89a"
        assertVersionAfterSave(im, b"GIF89a")

        # Test that adding a GIF89a feature changes the version
        im.info["transparency"] = 1
        assertVersionAfterSave(im, b"GIF89a")

        # Test that a GIF87a image is also saved in that format
        with Image.open("Tests/images/test.colors.gif") as im:
            assertVersionAfterSave(im, b"GIF87a")

            # Test that a GIF89a image is also saved in that format
            im.info["version"] = b"GIF89a"
            assertVersionAfterSave(im, b"GIF87a")

    def test_append_images(self):
        out = self.tempfile("temp.gif")

        # Test appending single frame images
        im = Image.new("RGB", (100, 100), "#f00")
        ims = [Image.new("RGB", (100, 100), color) for color in ["#0f0", "#00f"]]
        im.copy().save(out, save_all=True, append_images=ims)

        with Image.open(out) as reread:
            self.assertEqual(reread.n_frames, 3)

        # Tests appending using a generator
        def imGenerator(ims):
            yield from ims

        im.save(out, save_all=True, append_images=imGenerator(ims))

        with Image.open(out) as reread:
            self.assertEqual(reread.n_frames, 3)

        # Tests appending single and multiple frame images
        with Image.open("Tests/images/dispose_none.gif") as im:
            with Image.open("Tests/images/dispose_prev.gif") as im2:
                im.save(out, save_all=True, append_images=[im2])

        with Image.open(out) as reread:
            self.assertEqual(reread.n_frames, 10)

    def test_transparent_optimize(self):
        # from issue #2195, if the transparent color is incorrectly
        # optimized out, gif loses transparency
        # Need a palette that isn't using the 0 color, and one
        # that's > 128 items where the transparent color is actually
        # the top palette entry to trigger the bug.

        data = bytes(range(1, 254))
        palette = ImagePalette.ImagePalette("RGB", list(range(256)) * 3)

        im = Image.new("L", (253, 1))
        im.frombytes(data)
        im.putpalette(palette)

        out = self.tempfile("temp.gif")
        im.save(out, transparency=253)
        with Image.open(out) as reloaded:

            self.assertEqual(reloaded.info["transparency"], 253)

    def test_rgb_transparency(self):
        out = self.tempfile("temp.gif")

        # Single frame
        im = Image.new("RGB", (1, 1))
        im.info["transparency"] = (255, 0, 0)
        self.assert_warning(UserWarning, im.save, out)

        with Image.open(out) as reloaded:
            self.assertNotIn("transparency", reloaded.info)

        # Multiple frames
        im = Image.new("RGB", (1, 1))
        im.info["transparency"] = b""
        ims = [Image.new("RGB", (1, 1))]
        self.assert_warning(UserWarning, im.save, out, save_all=True, append_images=ims)

        with Image.open(out) as reloaded:
            self.assertNotIn("transparency", reloaded.info)

    def test_bbox(self):
        out = self.tempfile("temp.gif")

        im = Image.new("RGB", (100, 100), "#fff")
        ims = [Image.new("RGB", (100, 100), "#000")]
        im.save(out, save_all=True, append_images=ims)

        with Image.open(out) as reread:
            self.assertEqual(reread.n_frames, 2)

    def test_palette_save_L(self):
        # generate an L mode image with a separate palette

        im = hopper("P")
        im_l = Image.frombytes("L", im.size, im.tobytes())
        palette = bytes(im.getpalette())

        out = self.tempfile("temp.gif")
        im_l.save(out, palette=palette)

        with Image.open(out) as reloaded:

            self.assert_image_equal(reloaded.convert("RGB"), im.convert("RGB"))

    def test_palette_save_P(self):
        # pass in a different palette, then construct what the image
        # would look like.
        # Forcing a non-straight grayscale palette.

        im = hopper("P")
        palette = bytes([255 - i // 3 for i in range(768)])

        out = self.tempfile("temp.gif")
        im.save(out, palette=palette)

        with Image.open(out) as reloaded:
            im.putpalette(palette)
            self.assert_image_equal(reloaded, im)

    def test_palette_save_ImagePalette(self):
        # pass in a different palette, as an ImagePalette.ImagePalette
        # effectively the same as test_palette_save_P

        im = hopper("P")
        palette = ImagePalette.ImagePalette("RGB", list(range(256))[::-1] * 3)

        out = self.tempfile("temp.gif")
        im.save(out, palette=palette)

        with Image.open(out) as reloaded:
            im.putpalette(palette)
            self.assert_image_equal(reloaded, im)

    def test_save_I(self):
        # Test saving something that would trigger the auto-convert to 'L'

        im = hopper("I")

        out = self.tempfile("temp.gif")
        im.save(out)

        with Image.open(out) as reloaded:
            self.assert_image_equal(reloaded.convert("L"), im.convert("L"))

    def test_getdata(self):
        # test getheader/getdata against legacy values
        # Create a 'P' image with holes in the palette
        im = Image._wedge().resize((16, 16), Image.NEAREST)
        im.putpalette(ImagePalette.ImagePalette("RGB"))
        im.info = {"background": 0}

        passed_palette = bytes([255 - i // 3 for i in range(768)])

        GifImagePlugin._FORCE_OPTIMIZE = True
        try:
            h = GifImagePlugin.getheader(im, passed_palette)
            d = GifImagePlugin.getdata(im)

            import pickle

            # Enable to get target values on pre-refactor version
            # with open('Tests/images/gif_header_data.pkl', 'wb') as f:
            #    pickle.dump((h, d), f, 1)
            with open("Tests/images/gif_header_data.pkl", "rb") as f:
                (h_target, d_target) = pickle.load(f)

            self.assertEqual(h, h_target)
            self.assertEqual(d, d_target)
        finally:
            GifImagePlugin._FORCE_OPTIMIZE = False

    def test_lzw_bits(self):
        # see https://github.com/python-pillow/Pillow/issues/2811
        with Image.open("Tests/images/issue_2811.gif") as im:
            self.assertEqual(im.tile[0][3][0], 11)  # LZW bits
            # codec error prepatch
            im.load()

    def test_extents(self):
        with Image.open("Tests/images/test_extents.gif") as im:
            self.assertEqual(im.size, (100, 100))
            im.seek(1)
            self.assertEqual(im.size, (150, 150))



import subprocess
import sys

from .helper import PillowTestCase

try:
    from PIL import ImageGrab

    class TestImageGrab(PillowTestCase):
        def test_grab(self):
            for im in [
                ImageGrab.grab(),
                ImageGrab.grab(include_layered_windows=True),
                ImageGrab.grab(all_screens=True),
            ]:
                self.assert_image(im, im.mode, im.size)

        def test_grabclipboard(self):
            if sys.platform == "darwin":
                subprocess.call(["screencapture", "-cx"])
            else:
                p = subprocess.Popen(
                    ["powershell", "-command", "-"], stdin=subprocess.PIPE
                )
                p.stdin.write(
                    b"""[Reflection.Assembly]::LoadWithPartialName("System.Drawing")
[Reflection.Assembly]::LoadWithPartialName("System.Windows.Forms")
$bmp = New-Object Drawing.Bitmap 200, 200
[Windows.Forms.Clipboard]::SetImage($bmp)"""
                )
                p.communicate()

            im = ImageGrab.grabclipboard()
            self.assert_image(im, im.mode, im.size)


except ImportError:

    class TestImageGrab(PillowTestCase):
        def test_skip(self):
            self.skipTest("ImportError")


class TestImageGrabImport(PillowTestCase):
    def test_import(self):
        # Arrange
        exception = None

        # Act
        try:
            from PIL import ImageGrab

            ImageGrab.__name__  # dummy to prevent Pyflakes warning
        except Exception as e:
            exception = e

        # Assert
        if sys.platform in ["win32", "darwin"]:
            self.assertIsNone(exception)
        else:
            self.assertIsInstance(exception, ImportError)
            self.assertEqual(str(exception), "ImageGrab is macOS and Windows only")



from PIL.Image import (
    FLIP_LEFT_RIGHT,
    FLIP_TOP_BOTTOM,
    ROTATE_90,
    ROTATE_180,
    ROTATE_270,
    TRANSPOSE,
    TRANSVERSE,
)

from . import helper
from .helper import PillowTestCase


class TestImageTranspose(PillowTestCase):

    hopper = {
        mode: helper.hopper(mode).crop((0, 0, 121, 127)).copy()
        for mode in ["L", "RGB", "I;16", "I;16L", "I;16B"]
    }

    def test_flip_left_right(self):
        def transpose(mode):
            im = self.hopper[mode]
            out = im.transpose(FLIP_LEFT_RIGHT)
            self.assertEqual(out.mode, mode)
            self.assertEqual(out.size, im.size)

            x, y = im.size
            self.assertEqual(im.getpixel((1, 1)), out.getpixel((x - 2, 1)))
            self.assertEqual(im.getpixel((x - 2, 1)), out.getpixel((1, 1)))
            self.assertEqual(im.getpixel((1, y - 2)), out.getpixel((x - 2, y - 2)))
            self.assertEqual(im.getpixel((x - 2, y - 2)), out.getpixel((1, y - 2)))

        for mode in self.hopper:
            transpose(mode)

    def test_flip_top_bottom(self):
        def transpose(mode):
            im = self.hopper[mode]
            out = im.transpose(FLIP_TOP_BOTTOM)
            self.assertEqual(out.mode, mode)
            self.assertEqual(out.size, im.size)

            x, y = im.size
            self.assertEqual(im.getpixel((1, 1)), out.getpixel((1, y - 2)))
            self.assertEqual(im.getpixel((x - 2, 1)), out.getpixel((x - 2, y - 2)))
            self.assertEqual(im.getpixel((1, y - 2)), out.getpixel((1, 1)))
            self.assertEqual(im.getpixel((x - 2, y - 2)), out.getpixel((x - 2, 1)))

        for mode in self.hopper:
            transpose(mode)

    def test_rotate_90(self):
        def transpose(mode):
            im = self.hopper[mode]
            out = im.transpose(ROTATE_90)
            self.assertEqual(out.mode, mode)
            self.assertEqual(out.size, im.size[::-1])

            x, y = im.size
            self.assertEqual(im.getpixel((1, 1)), out.getpixel((1, x - 2)))
            self.assertEqual(im.getpixel((x - 2, 1)), out.getpixel((1, 1)))
            self.assertEqual(im.getpixel((1, y - 2)), out.getpixel((y - 2, x - 2)))
            self.assertEqual(im.getpixel((x - 2, y - 2)), out.getpixel((y - 2, 1)))

        for mode in self.hopper:
            transpose(mode)

    def test_rotate_180(self):
        def transpose(mode):
            im = self.hopper[mode]
            out = im.transpose(ROTATE_180)
            self.assertEqual(out.mode, mode)
            self.assertEqual(out.size, im.size)

            x, y = im.size
            self.assertEqual(im.getpixel((1, 1)), out.getpixel((x - 2, y - 2)))
            self.assertEqual(im.getpixel((x - 2, 1)), out.getpixel((1, y - 2)))
            self.assertEqual(im.getpixel((1, y - 2)), out.getpixel((x - 2, 1)))
            self.assertEqual(im.getpixel((x - 2, y - 2)), out.getpixel((1, 1)))

        for mode in self.hopper:
            transpose(mode)

    def test_rotate_270(self):
        def transpose(mode):
            im = self.hopper[mode]
            out = im.transpose(ROTATE_270)
            self.assertEqual(out.mode, mode)
            self.assertEqual(out.size, im.size[::-1])

            x, y = im.size
            self.assertEqual(im.getpixel((1, 1)), out.getpixel((y - 2, 1)))
            self.assertEqual(im.getpixel((x - 2, 1)), out.getpixel((y - 2, x - 2)))
            self.assertEqual(im.getpixel((1, y - 2)), out.getpixel((1, 1)))
            self.assertEqual(im.getpixel((x - 2, y - 2)), out.getpixel((1, x - 2)))

        for mode in self.hopper:
            transpose(mode)

    def test_transpose(self):
        def transpose(mode):
            im = self.hopper[mode]
            out = im.transpose(TRANSPOSE)
            self.assertEqual(out.mode, mode)
            self.assertEqual(out.size, im.size[::-1])

            x, y = im.size
            self.assertEqual(im.getpixel((1, 1)), out.getpixel((1, 1)))
            self.assertEqual(im.getpixel((x - 2, 1)), out.getpixel((1, x - 2)))
            self.assertEqual(im.getpixel((1, y - 2)), out.getpixel((y - 2, 1)))
            self.assertEqual(im.getpixel((x - 2, y - 2)), out.getpixel((y - 2, x - 2)))

        for mode in self.hopper:
            transpose(mode)

    def test_tranverse(self):
        def transpose(mode):
            im = self.hopper[mode]
            out = im.transpose(TRANSVERSE)
            self.assertEqual(out.mode, mode)
            self.assertEqual(out.size, im.size[::-1])

            x, y = im.size
            self.assertEqual(im.getpixel((1, 1)), out.getpixel((y - 2, x - 2)))
            self.assertEqual(im.getpixel((x - 2, 1)), out.getpixel((y - 2, 1)))
            self.assertEqual(im.getpixel((1, y - 2)), out.getpixel((1, x - 2)))
            self.assertEqual(im.getpixel((x - 2, y - 2)), out.getpixel((1, 1)))

        for mode in self.hopper:
            transpose(mode)

    def test_roundtrip(self):
        for mode in self.hopper:
            im = self.hopper[mode]

            def transpose(first, second):
                return im.transpose(first).transpose(second)

            self.assert_image_equal(im, transpose(FLIP_LEFT_RIGHT, FLIP_LEFT_RIGHT))
            self.assert_image_equal(im, transpose(FLIP_TOP_BOTTOM, FLIP_TOP_BOTTOM))
            self.assert_image_equal(im, transpose(ROTATE_90, ROTATE_270))
            self.assert_image_equal(im, transpose(ROTATE_180, ROTATE_180))
            self.assert_image_equal(
                im.transpose(TRANSPOSE), transpose(ROTATE_90, FLIP_TOP_BOTTOM)
            )
            self.assert_image_equal(
                im.transpose(TRANSPOSE), transpose(ROTATE_270, FLIP_LEFT_RIGHT)
            )
            self.assert_image_equal(
                im.transpose(TRANSVERSE), transpose(ROTATE_90, FLIP_LEFT_RIGHT)
            )
            self.assert_image_equal(
                im.transpose(TRANSVERSE), transpose(ROTATE_270, FLIP_TOP_BOTTOM)
            )
            self.assert_image_equal(
                im.transpose(TRANSVERSE), transpose(ROTATE_180, TRANSPOSE)
            )



import io

from PIL import BmpImagePlugin, Image

from .helper import PillowTestCase, hopper


class TestFileBmp(PillowTestCase):
    def roundtrip(self, im):
        outfile = self.tempfile("temp.bmp")

        im.save(outfile, "BMP")

        with Image.open(outfile) as reloaded:
            reloaded.load()
            self.assertEqual(im.mode, reloaded.mode)
            self.assertEqual(im.size, reloaded.size)
            self.assertEqual(reloaded.format, "BMP")
            self.assertEqual(reloaded.get_format_mimetype(), "image/bmp")

    def test_sanity(self):
        self.roundtrip(hopper())

        self.roundtrip(hopper("1"))
        self.roundtrip(hopper("L"))
        self.roundtrip(hopper("P"))
        self.roundtrip(hopper("RGB"))

    def test_invalid_file(self):
        with open("Tests/images/flower.jpg", "rb") as fp:
            self.assertRaises(SyntaxError, BmpImagePlugin.BmpImageFile, fp)

    def test_save_to_bytes(self):
        output = io.BytesIO()
        im = hopper()
        im.save(output, "BMP")

        output.seek(0)
        with Image.open(output) as reloaded:
            self.assertEqual(im.mode, reloaded.mode)
            self.assertEqual(im.size, reloaded.size)
            self.assertEqual(reloaded.format, "BMP")

    def test_save_too_large(self):
        outfile = self.tempfile("temp.bmp")
        with Image.new("RGB", (1, 1)) as im:
            im._size = (37838, 37838)
            with self.assertRaises(ValueError):
                im.save(outfile)

    def test_dpi(self):
        dpi = (72, 72)

        output = io.BytesIO()
        with hopper() as im:
            im.save(output, "BMP", dpi=dpi)

        output.seek(0)
        with Image.open(output) as reloaded:
            self.assertEqual(reloaded.info["dpi"], dpi)

    def test_save_bmp_with_dpi(self):
        # Test for #1301
        # Arrange
        outfile = self.tempfile("temp.jpg")
        with Image.open("Tests/images/hopper.bmp") as im:

            # Act
            im.save(outfile, "JPEG", dpi=im.info["dpi"])

            # Assert
            with Image.open(outfile) as reloaded:
                reloaded.load()
                self.assertEqual(im.info["dpi"], reloaded.info["dpi"])
                self.assertEqual(im.size, reloaded.size)
                self.assertEqual(reloaded.format, "JPEG")

    def test_load_dpi_rounding(self):
        # Round up
        with Image.open("Tests/images/hopper.bmp") as im:
            self.assertEqual(im.info["dpi"], (96, 96))

        # Round down
        with Image.open("Tests/images/hopper_roundDown.bmp") as im:
            self.assertEqual(im.info["dpi"], (72, 72))

    def test_save_dpi_rounding(self):
        outfile = self.tempfile("temp.bmp")
        with Image.open("Tests/images/hopper.bmp") as im:
            im.save(outfile, dpi=(72.2, 72.2))
            with Image.open(outfile) as reloaded:
                self.assertEqual(reloaded.info["dpi"], (72, 72))

            im.save(outfile, dpi=(72.8, 72.8))
        with Image.open(outfile) as reloaded:
            self.assertEqual(reloaded.info["dpi"], (73, 73))

    def test_load_dib(self):
        # test for #1293, Imagegrab returning Unsupported Bitfields Format
        with Image.open("Tests/images/clipboard.dib") as im:
            self.assertEqual(im.format, "DIB")
            self.assertEqual(im.get_format_mimetype(), "image/bmp")

            with Image.open("Tests/images/clipboard_target.png") as target:
                self.assert_image_equal(im, target)

    def test_save_dib(self):
        outfile = self.tempfile("temp.dib")

        with Image.open("Tests/images/clipboard.dib") as im:
            im.save(outfile)

            with Image.open(outfile) as reloaded:
                self.assertEqual(reloaded.format, "DIB")
                self.assertEqual(reloaded.get_format_mimetype(), "image/bmp")
                self.assert_image_equal(im, reloaded)

    def test_rgba_bitfields(self):
        # This test image has been manually hexedited
        # to change the bitfield compression in the header from XBGR to RGBA
        with Image.open("Tests/images/rgb32bf-rgba.bmp") as im:

            # So before the comparing the image, swap the channels
            b, g, r = im.split()[1:]
            im = Image.merge("RGB", (r, g, b))

        with Image.open("Tests/images/bmp/q/rgb32bf-xbgr.bmp") as target:
            self.assert_image_equal(im, target)



from PIL import Image

from .helper import PillowTestCase, fromstring, tostring


class TestImageDraft(PillowTestCase):
    def setUp(self):
        codecs = dir(Image.core)
        if "jpeg_encoder" not in codecs or "jpeg_decoder" not in codecs:
            self.skipTest("jpeg support not available")

    def draft_roundtrip(self, in_mode, in_size, req_mode, req_size):
        im = Image.new(in_mode, in_size)
        data = tostring(im, "JPEG")
        im = fromstring(data)
        mode, box = im.draft(req_mode, req_size)
        scale, _ = im.decoderconfig
        self.assertEqual(box[:2], (0, 0))
        self.assertTrue((im.width - scale) < box[2] <= im.width)
        self.assertTrue((im.height - scale) < box[3] <= im.height)
        return im

    def test_size(self):
        for in_size, req_size, out_size in [
            ((435, 361), (2048, 2048), (435, 361)),  # bigger
            ((435, 361), (435, 361), (435, 361)),  # same
            ((128, 128), (64, 64), (64, 64)),
            ((128, 128), (32, 32), (32, 32)),
            ((128, 128), (16, 16), (16, 16)),
            # large requested width
            ((435, 361), (218, 128), (435, 361)),  # almost 2x
            ((435, 361), (217, 128), (218, 181)),  # more than 2x
            ((435, 361), (109, 64), (218, 181)),  # almost 4x
            ((435, 361), (108, 64), (109, 91)),  # more than 4x
            ((435, 361), (55, 32), (109, 91)),  # almost 8x
            ((435, 361), (54, 32), (55, 46)),  # more than 8x
            ((435, 361), (27, 16), (55, 46)),  # more than 16x
            # and vice versa
            ((435, 361), (128, 181), (435, 361)),  # almost 2x
            ((435, 361), (128, 180), (218, 181)),  # more than 2x
            ((435, 361), (64, 91), (218, 181)),  # almost 4x
            ((435, 361), (64, 90), (109, 91)),  # more than 4x
            ((435, 361), (32, 46), (109, 91)),  # almost 8x
            ((435, 361), (32, 45), (55, 46)),  # more than 8x
            ((435, 361), (16, 22), (55, 46)),  # more than 16x
        ]:
            im = self.draft_roundtrip("L", in_size, None, req_size)
            im.load()
            self.assertEqual(im.size, out_size)

    def test_mode(self):
        for in_mode, req_mode, out_mode in [
            ("RGB", "1", "RGB"),
            ("RGB", "L", "L"),
            ("RGB", "RGB", "RGB"),
            ("RGB", "YCbCr", "YCbCr"),
            ("L", "1", "L"),
            ("L", "L", "L"),
            ("L", "RGB", "L"),
            ("L", "YCbCr", "L"),
            ("CMYK", "1", "CMYK"),
            ("CMYK", "L", "CMYK"),
            ("CMYK", "RGB", "CMYK"),
            ("CMYK", "YCbCr", "CMYK"),
        ]:
            im = self.draft_roundtrip(in_mode, (64, 64), req_mode, None)
            im.load()
            self.assertEqual(im.mode, out_mode)

    def test_several_drafts(self):
        im = self.draft_roundtrip("L", (128, 128), None, (64, 64))
        im.draft(None, (64, 64))
        im.load()



from PIL import Image

from .helper import PillowTestCase, hopper

try:
    from PIL import _webp

    HAVE_WEBP = True
except ImportError:
    HAVE_WEBP = False


class TestFileWebpLossless(PillowTestCase):
    def setUp(self):
        if not HAVE_WEBP:
            self.skipTest("WebP support not installed")
            return

        if _webp.WebPDecoderVersion() < 0x0200:
            self.skipTest("lossless not included")

        self.rgb_mode = "RGB"

    def test_write_lossless_rgb(self):
        temp_file = self.tempfile("temp.webp")

        hopper(self.rgb_mode).save(temp_file, lossless=True)

        with Image.open(temp_file) as image:
            image.load()

            self.assertEqual(image.mode, self.rgb_mode)
            self.assertEqual(image.size, (128, 128))
            self.assertEqual(image.format, "WEBP")
            image.load()
            image.getdata()

            self.assert_image_equal(image, hopper(self.rgb_mode))



import os
from glob import glob
from itertools import product

from PIL import Image

from .helper import PillowTestCase

_TGA_DIR = os.path.join("Tests", "images", "tga")
_TGA_DIR_COMMON = os.path.join(_TGA_DIR, "common")


class TestFileTga(PillowTestCase):

    _MODES = ("L", "LA", "P", "RGB", "RGBA")
    _ORIGINS = ("tl", "bl")

    _ORIGIN_TO_ORIENTATION = {"tl": 1, "bl": -1}

    def test_sanity(self):
        for mode in self._MODES:
            png_paths = glob(
                os.path.join(_TGA_DIR_COMMON, "*x*_{}.png".format(mode.lower()))
            )

            for png_path in png_paths:
                with Image.open(png_path) as reference_im:
                    self.assertEqual(reference_im.mode, mode)

                    path_no_ext = os.path.splitext(png_path)[0]
                    for origin, rle in product(self._ORIGINS, (True, False)):
                        tga_path = "{}_{}_{}.tga".format(
                            path_no_ext, origin, "rle" if rle else "raw"
                        )

                        with Image.open(tga_path) as original_im:
                            self.assertEqual(original_im.format, "TGA")
                            self.assertEqual(
                                original_im.get_format_mimetype(), "image/x-tga"
                            )
                            if rle:
                                self.assertEqual(
                                    original_im.info["compression"], "tga_rle"
                                )
                            self.assertEqual(
                                original_im.info["orientation"],
                                self._ORIGIN_TO_ORIENTATION[origin],
                            )
                            if mode == "P":
                                self.assertEqual(
                                    original_im.getpalette(), reference_im.getpalette()
                                )

                            self.assert_image_equal(original_im, reference_im)

                            # Generate a new test name every time so the
                            # test will not fail with permission error
                            # on Windows.
                            out = self.tempfile("temp.tga")

                            original_im.save(out, rle=rle)
                            with Image.open(out) as saved_im:
                                if rle:
                                    self.assertEqual(
                                        saved_im.info["compression"],
                                        original_im.info["compression"],
                                    )
                                self.assertEqual(
                                    saved_im.info["orientation"],
                                    original_im.info["orientation"],
                                )
                                if mode == "P":
                                    self.assertEqual(
                                        saved_im.getpalette(), original_im.getpalette()
                                    )

                                self.assert_image_equal(saved_im, original_im)

    def test_id_field(self):
        # tga file with id field
        test_file = "Tests/images/tga_id_field.tga"

        # Act
        with Image.open(test_file) as im:

            # Assert
            self.assertEqual(im.size, (100, 100))

    def test_id_field_rle(self):
        # tga file with id field
        test_file = "Tests/images/rgb32rle.tga"

        # Act
        with Image.open(test_file) as im:

            # Assert
            self.assertEqual(im.size, (199, 199))

    def test_save(self):
        test_file = "Tests/images/tga_id_field.tga"
        with Image.open(test_file) as im:
            out = self.tempfile("temp.tga")

            # Save
            im.save(out)
            with Image.open(out) as test_im:
                self.assertEqual(test_im.size, (100, 100))
                self.assertEqual(test_im.info["id_section"], im.info["id_section"])

            # RGBA save
            im.convert("RGBA").save(out)
        with Image.open(out) as test_im:
            self.assertEqual(test_im.size, (100, 100))

    def test_save_id_section(self):
        test_file = "Tests/images/rgb32rle.tga"
        with Image.open(test_file) as im:
            out = self.tempfile("temp.tga")

            # Check there is no id section
            im.save(out)
        with Image.open(out) as test_im:
            self.assertNotIn("id_section", test_im.info)

        # Save with custom id section
        im.save(out, id_section=b"Test content")
        with Image.open(out) as test_im:
            self.assertEqual(test_im.info["id_section"], b"Test content")

        # Save with custom id section greater than 255 characters
        id_section = b"Test content" * 25
        self.assert_warning(UserWarning, lambda: im.save(out, id_section=id_section))
        with Image.open(out) as test_im:
            self.assertEqual(test_im.info["id_section"], id_section[:255])

        test_file = "Tests/images/tga_id_field.tga"
        with Image.open(test_file) as im:

            # Save with no id section
            im.save(out, id_section="")
        with Image.open(out) as test_im:
            self.assertNotIn("id_section", test_im.info)

    def test_save_orientation(self):
        test_file = "Tests/images/rgb32rle.tga"
        out = self.tempfile("temp.tga")
        with Image.open(test_file) as im:
            self.assertEqual(im.info["orientation"], -1)

            im.save(out, orientation=1)
        with Image.open(out) as test_im:
            self.assertEqual(test_im.info["orientation"], 1)

    def test_save_rle(self):
        test_file = "Tests/images/rgb32rle.tga"
        with Image.open(test_file) as im:
            self.assertEqual(im.info["compression"], "tga_rle")

            out = self.tempfile("temp.tga")

            # Save
            im.save(out)
        with Image.open(out) as test_im:
            self.assertEqual(test_im.size, (199, 199))
            self.assertEqual(test_im.info["compression"], "tga_rle")

        # Save without compression
        im.save(out, compression=None)
        with Image.open(out) as test_im:
            self.assertNotIn("compression", test_im.info)

        # RGBA save
        im.convert("RGBA").save(out)
        with Image.open(out) as test_im:
            self.assertEqual(test_im.size, (199, 199))

        test_file = "Tests/images/tga_id_field.tga"
        with Image.open(test_file) as im:
            self.assertNotIn("compression", im.info)

            # Save with compression
            im.save(out, compression="tga_rle")
        with Image.open(out) as test_im:
            self.assertEqual(test_im.info["compression"], "tga_rle")

    def test_save_l_transparency(self):
        # There are 559 transparent pixels in la.tga.
        num_transparent = 559

        in_file = "Tests/images/la.tga"
        with Image.open(in_file) as im:
            self.assertEqual(im.mode, "LA")
            self.assertEqual(im.getchannel("A").getcolors()[0][0], num_transparent)

            out = self.tempfile("temp.tga")
            im.save(out)

        with Image.open(out) as test_im:
            self.assertEqual(test_im.mode, "LA")
            self.assertEqual(test_im.getchannel("A").getcolors()[0][0], num_transparent)

            self.assert_image_equal(im, test_im)



from PIL import Image, ImageChops

from .helper import PillowTestCase, hopper

BLACK = (0, 0, 0)
BROWN = (127, 64, 0)
CYAN = (0, 255, 255)
DARK_GREEN = (0, 128, 0)
GREEN = (0, 255, 0)
ORANGE = (255, 128, 0)
WHITE = (255, 255, 255)

GREY = 128


class TestImageChops(PillowTestCase):
    def test_sanity(self):

        im = hopper("L")

        ImageChops.constant(im, 128)
        ImageChops.duplicate(im)
        ImageChops.invert(im)
        ImageChops.lighter(im, im)
        ImageChops.darker(im, im)
        ImageChops.difference(im, im)
        ImageChops.multiply(im, im)
        ImageChops.screen(im, im)

        ImageChops.add(im, im)
        ImageChops.add(im, im, 2.0)
        ImageChops.add(im, im, 2.0, 128)
        ImageChops.subtract(im, im)
        ImageChops.subtract(im, im, 2.0)
        ImageChops.subtract(im, im, 2.0, 128)

        ImageChops.add_modulo(im, im)
        ImageChops.subtract_modulo(im, im)

        ImageChops.blend(im, im, 0.5)
        ImageChops.composite(im, im, im)

        ImageChops.offset(im, 10)
        ImageChops.offset(im, 10, 20)

    def test_add(self):
        # Arrange
        with Image.open("Tests/images/imagedraw_ellipse_RGB.png") as im1:
            with Image.open("Tests/images/imagedraw_floodfill_RGB.png") as im2:

                # Act
                new = ImageChops.add(im1, im2)

        # Assert
        self.assertEqual(new.getbbox(), (25, 25, 76, 76))
        self.assertEqual(new.getpixel((50, 50)), ORANGE)

    def test_add_scale_offset(self):
        # Arrange
        with Image.open("Tests/images/imagedraw_ellipse_RGB.png") as im1:
            with Image.open("Tests/images/imagedraw_floodfill_RGB.png") as im2:

                # Act
                new = ImageChops.add(im1, im2, scale=2.5, offset=100)

        # Assert
        self.assertEqual(new.getbbox(), (0, 0, 100, 100))
        self.assertEqual(new.getpixel((50, 50)), (202, 151, 100))

    def test_add_clip(self):
        # Arrange
        im = hopper()

        # Act
        new = ImageChops.add(im, im)

        # Assert
        self.assertEqual(new.getpixel((50, 50)), (255, 255, 254))

    def test_add_modulo(self):
        # Arrange
        with Image.open("Tests/images/imagedraw_ellipse_RGB.png") as im1:
            with Image.open("Tests/images/imagedraw_floodfill_RGB.png") as im2:

                # Act
                new = ImageChops.add_modulo(im1, im2)

        # Assert
        self.assertEqual(new.getbbox(), (25, 25, 76, 76))
        self.assertEqual(new.getpixel((50, 50)), ORANGE)

    def test_add_modulo_no_clip(self):
        # Arrange
        im = hopper()

        # Act
        new = ImageChops.add_modulo(im, im)

        # Assert
        self.assertEqual(new.getpixel((50, 50)), (224, 76, 254))

    def test_blend(self):
        # Arrange
        with Image.open("Tests/images/imagedraw_ellipse_RGB.png") as im1:
            with Image.open("Tests/images/imagedraw_floodfill_RGB.png") as im2:

                # Act
                new = ImageChops.blend(im1, im2, 0.5)

        # Assert
        self.assertEqual(new.getbbox(), (25, 25, 76, 76))
        self.assertEqual(new.getpixel((50, 50)), BROWN)

    def test_constant(self):
        # Arrange
        im = Image.new("RGB", (20, 10))

        # Act
        new = ImageChops.constant(im, GREY)

        # Assert
        self.assertEqual(new.size, im.size)
        self.assertEqual(new.getpixel((0, 0)), GREY)
        self.assertEqual(new.getpixel((19, 9)), GREY)

    def test_darker_image(self):
        # Arrange
        with Image.open("Tests/images/imagedraw_chord_RGB.png") as im1:
            with Image.open("Tests/images/imagedraw_outline_chord_RGB.png") as im2:

                # Act
                new = ImageChops.darker(im1, im2)

                # Assert
                self.assert_image_equal(new, im2)

    def test_darker_pixel(self):
        # Arrange
        im1 = hopper()
        with Image.open("Tests/images/imagedraw_chord_RGB.png") as im2:

            # Act
            new = ImageChops.darker(im1, im2)

        # Assert
        self.assertEqual(new.getpixel((50, 50)), (240, 166, 0))

    def test_difference(self):
        # Arrange
        with Image.open("Tests/images/imagedraw_arc_end_le_start.png") as im1:
            with Image.open("Tests/images/imagedraw_arc_no_loops.png") as im2:

                # Act
                new = ImageChops.difference(im1, im2)

        # Assert
        self.assertEqual(new.getbbox(), (25, 25, 76, 76))

    def test_difference_pixel(self):
        # Arrange
        im1 = hopper()
        with Image.open("Tests/images/imagedraw_polygon_kite_RGB.png") as im2:

            # Act
            new = ImageChops.difference(im1, im2)

        # Assert
        self.assertEqual(new.getpixel((50, 50)), (240, 166, 128))

    def test_duplicate(self):
        # Arrange
        im = hopper()

        # Act
        new = ImageChops.duplicate(im)

        # Assert
        self.assert_image_equal(new, im)

    def test_invert(self):
        # Arrange
        with Image.open("Tests/images/imagedraw_floodfill_RGB.png") as im:

            # Act
            new = ImageChops.invert(im)

        # Assert
        self.assertEqual(new.getbbox(), (0, 0, 100, 100))
        self.assertEqual(new.getpixel((0, 0)), WHITE)
        self.assertEqual(new.getpixel((50, 50)), CYAN)

    def test_lighter_image(self):
        # Arrange
        with Image.open("Tests/images/imagedraw_chord_RGB.png") as im1:
            with Image.open("Tests/images/imagedraw_outline_chord_RGB.png") as im2:

                # Act
                new = ImageChops.lighter(im1, im2)

        # Assert
        self.assert_image_equal(new, im1)

    def test_lighter_pixel(self):
        # Arrange
        im1 = hopper()
        with Image.open("Tests/images/imagedraw_chord_RGB.png") as im2:

            # Act
            new = ImageChops.lighter(im1, im2)

        # Assert
        self.assertEqual(new.getpixel((50, 50)), (255, 255, 127))

    def test_multiply_black(self):
        """If you multiply an image with a solid black image,
        the result is black."""
        # Arrange
        im1 = hopper()
        black = Image.new("RGB", im1.size, "black")

        # Act
        new = ImageChops.multiply(im1, black)

        # Assert
        self.assert_image_equal(new, black)

    def test_multiply_green(self):
        # Arrange
        with Image.open("Tests/images/imagedraw_floodfill_RGB.png") as im:
            green = Image.new("RGB", im.size, "green")

            # Act
            new = ImageChops.multiply(im, green)

        # Assert
        self.assertEqual(new.getbbox(), (25, 25, 76, 76))
        self.assertEqual(new.getpixel((25, 25)), DARK_GREEN)
        self.assertEqual(new.getpixel((50, 50)), BLACK)

    def test_multiply_white(self):
        """If you multiply with a solid white image,
        the image is unaffected."""
        # Arrange
        im1 = hopper()
        white = Image.new("RGB", im1.size, "white")

        # Act
        new = ImageChops.multiply(im1, white)

        # Assert
        self.assert_image_equal(new, im1)

    def test_offset(self):
        # Arrange
        xoffset = 45
        yoffset = 20
        with Image.open("Tests/images/imagedraw_ellipse_RGB.png") as im:

            # Act
            new = ImageChops.offset(im, xoffset, yoffset)

        # Assert
        self.assertEqual(new.getbbox(), (0, 45, 100, 96))
        self.assertEqual(new.getpixel((50, 50)), BLACK)
        self.assertEqual(new.getpixel((50 + xoffset, 50 + yoffset)), DARK_GREEN)

        # Test no yoffset
        self.assertEqual(
            ImageChops.offset(im, xoffset), ImageChops.offset(im, xoffset, xoffset)
        )

    def test_screen(self):
        # Arrange
        with Image.open("Tests/images/imagedraw_ellipse_RGB.png") as im1:
            with Image.open("Tests/images/imagedraw_floodfill_RGB.png") as im2:

                # Act
                new = ImageChops.screen(im1, im2)

        # Assert
        self.assertEqual(new.getbbox(), (25, 25, 76, 76))
        self.assertEqual(new.getpixel((50, 50)), ORANGE)

    def test_subtract(self):
        # Arrange
        with Image.open("Tests/images/imagedraw_chord_RGB.png") as im1:
            with Image.open("Tests/images/imagedraw_outline_chord_RGB.png") as im2:

                # Act
                new = ImageChops.subtract(im1, im2)

        # Assert
        self.assertEqual(new.getbbox(), (25, 50, 76, 76))
        self.assertEqual(new.getpixel((50, 50)), GREEN)
        self.assertEqual(new.getpixel((50, 51)), BLACK)

    def test_subtract_scale_offset(self):
        # Arrange
        with Image.open("Tests/images/imagedraw_chord_RGB.png") as im1:
            with Image.open("Tests/images/imagedraw_outline_chord_RGB.png") as im2:

                # Act
                new = ImageChops.subtract(im1, im2, scale=2.5, offset=100)

        # Assert
        self.assertEqual(new.getbbox(), (0, 0, 100, 100))
        self.assertEqual(new.getpixel((50, 50)), (100, 202, 100))

    def test_subtract_clip(self):
        # Arrange
        im1 = hopper()
        with Image.open("Tests/images/imagedraw_chord_RGB.png") as im2:

            # Act
            new = ImageChops.subtract(im1, im2)

        # Assert
        self.assertEqual(new.getpixel((50, 50)), (0, 0, 127))

    def test_subtract_modulo(self):
        # Arrange
        with Image.open("Tests/images/imagedraw_chord_RGB.png") as im1:
            with Image.open("Tests/images/imagedraw_outline_chord_RGB.png") as im2:

                # Act
                new = ImageChops.subtract_modulo(im1, im2)

        # Assert
        self.assertEqual(new.getbbox(), (25, 50, 76, 76))
        self.assertEqual(new.getpixel((50, 50)), GREEN)
        self.assertEqual(new.getpixel((50, 51)), BLACK)

    def test_subtract_modulo_no_clip(self):
        # Arrange
        im1 = hopper()
        with Image.open("Tests/images/imagedraw_chord_RGB.png") as im2:

            # Act
            new = ImageChops.subtract_modulo(im1, im2)

        # Assert
        self.assertEqual(new.getpixel((50, 50)), (241, 167, 127))

    def test_logical(self):
        def table(op, a, b):
            out = []
            for x in (a, b):
                imx = Image.new("1", (1, 1), x)
                for y in (a, b):
                    imy = Image.new("1", (1, 1), y)
                    out.append(op(imx, imy).getpixel((0, 0)))
            return tuple(out)

        self.assertEqual(table(ImageChops.logical_and, 0, 1), (0, 0, 0, 255))
        self.assertEqual(table(ImageChops.logical_or, 0, 1), (0, 255, 255, 255))
        self.assertEqual(table(ImageChops.logical_xor, 0, 1), (0, 255, 255, 0))

        self.assertEqual(table(ImageChops.logical_and, 0, 128), (0, 0, 0, 255))
        self.assertEqual(table(ImageChops.logical_or, 0, 128), (0, 255, 255, 255))
        self.assertEqual(table(ImageChops.logical_xor, 0, 128), (0, 255, 255, 0))

        self.assertEqual(table(ImageChops.logical_and, 0, 255), (0, 0, 0, 255))
        self.assertEqual(table(ImageChops.logical_or, 0, 255), (0, 255, 255, 255))
        self.assertEqual(table(ImageChops.logical_xor, 0, 255), (0, 255, 255, 0))
